/** \file srvquery.c
 * \brief query routines for the stock properties
 * \internal
 *
 */
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"
#include "almlib.h"
#include "hstlib.h"
#include "srvip.h"
#include "srvipx.h"
#include "ttycmd.h"

int getStringFromInputField(CONTRACT *con,char *dataIn,char *strbuf,int buflen)
{
  int len,fmt;
  if (con == NULL || strbuf == NULL || dataIn == NULL || buflen < 1) return argument_list_error;
  if ((len=(int)con->EqmSizeIn) == 0) return illegal_read_write;
  memset(strbuf,0,buflen);
  fmt = LFMT(con->EqmFormatIn);
  switch (fmt)
  {
    case CF_TEXT:
      len = MIN(len,buflen);
      if (len < 1) return illegal_read_write;
      strncpy(strbuf,dataIn,len);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      len = MIN(fmtsizeof(fmt),buflen);
      strncpy(strbuf,dataIn,len);
      break;
    default: 
      return illegal_format;
  }
  return 0;
}
int T_numQuery(CONTRACT *con,char *data,short num)
{
  if (con->EqmSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_SHORT: *(short *)data = num; break;
    case CF_LONG: *(UINT32 *)data = (UINT32)num; break;
    case CF_FLOAT: *(float *)data = (float)num; break;
    case CF_DOUBLE: *(double *)data = (double) num; break;
    default: return illegal_format;
  }
  return 0;
}
int taggedStructQuery(CONTRACT *con,char *dataIn,char *data)
{
  int cc,k;
  char tag[32];

  if (con->EqmSizeOut == 0) return illegal_read_write;
  if ((cc=getStringFromInputField(con,dataIn,tag,TAG_NAME_SIZE)) != 0) return cc;
  tag[8] = 0;
  k = (int)con->EqmSizeOut;
  cc = queryStructFormat(tag,(BYTE *)data,LFMT(con->EqmFormatOut),&k);
  if (k < (int)con->EqmSizeOut) con->EqmSizeOut = k;
  return cc;
}
int bitfieldStructQuery(CONTRACT *con,char *dataIn,char *data)
{
  int cc,k;
  char tag[32];

  if (con->EqmSizeOut == 0) return illegal_read_write;
  if ((cc=getStringFromInputField(con,dataIn,tag,TAG_NAME_SIZE)) != 0) return cc;
  tag[8] = 0;
  k = (int)con->EqmSizeOut;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME16I:
      cc = queryBitfieldFormat(NULL,tag,(NAME16I *)data,&k);
      break;
    default:
      return illegal_format;
  }
  if (k < (int)con->EqmSizeOut) con->EqmSizeOut = k;
  return cc;
}
int getServerAliasList(CONTRACT *con,char *dataIn,char *data)
{
  int len,i,j;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      len = MIN(fmtsizeof(con->EqmFormatOut),32);
      break;
    default:  return illegal_format;
  }
  for (i=0,j=0; i<(int)con->EqmSizeOut && j<nAliasTableEntries; i += 2,j++)
  {
     strncpy(((NAME32 *)data)[i].name,gAliasTable[j].name,len);
     strncpy(((NAME32 *)data)[i+1].name,gAliasTable[j].alias,len);
  }
  if (i < (int)con->EqmSizeOut) con->EqmSizeOut = i;
  return 0;
}
int ndeviceQuery(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el;
  short n;
/*
  // get this to work for an "ndevices for property" call !
  int pid;
  ExportPropertyListStruct *pl=NULL;
  if (con->EqmSizeIn > 0)
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_TEXT:
      case CF_NAME16:
      case CF_NAME32:
        if ((pid=GetPrpId(con->EqmName,dataIn)) < 0) break;
        pid = PROPERTY_HASH_INDEX(pid);
        for (pl=el->EqmProperty[pid]; pl != NULL && strncmp(pl->prpName,dataIn,32); pl=pl->next);
        break;
      default:  return illegal_format;
    }
  }
*/
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
/*
    if (pl != NULL && pl->numDevices > 0) n = pl->numDevices;
    else 
*/
    n = el->EqmNumDevices;
    return T_numQuery(con,data,n);
  }
  return non_existent_elem;
}
int deviceQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,k,slen,start;
  ExportListStruct *el;
  int cc = 0;
  char btok[DEVICE_NAME_SIZE], etok[DEVICE_NAME_SIZE], mtok[DEVICE_NAME_SIZE];
  int blen = 0, elen = 0, mlen = 0, checkWildCard = FALSE;

  dataIn = dataIn;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  start = con->EqmDeviceName[0] == '#' ? atoi(&con->EqmDeviceName[1])-1 : 0;
  if (start < 0) start = 0;

  if (strcmp(con->EqmDeviceName,"*") && strchr(con->EqmDeviceName,'*') != NULL)
  { /* device query with non-trivial wildcard string */
    getWildcardTokens(con->EqmDeviceName,btok,etok,mtok);
    blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
    checkWildCard = TRUE;
  }

  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME8: 
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      slen = fmtsizeof(con->EqmFormatOut); 
      break;
    default: 
      return illegal_format;
  }
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) return non_existent_elem;
  for (i=0,k=0; i<(int)con->EqmSizeOut && i+start<el->EqmNumDevices; i++)
  {
    if (el->EqmDeviceList && strlen(el->EqmDeviceList[i+start].name))
    {
      if (checkWildCard && 
          !wildcardMatches(el->EqmDeviceList[i+start].name,btok,blen,etok,elen,mtok,mlen))
        continue;
      strncpy(&data[k*slen],el->EqmDeviceList[i+start].name,slen);
    }
    else 
    {
      if (checkWildCard) continue;
      sprintf(&data[k*slen],"#%d",i);
      if (k == 0) k = 1;
      break;
    }
    k++;
  }
  if (el->hasPropertyDeviceNames) 
  {
    cc = has_query_function | CE_SENDDATA;
  }
  if (k < (int)con->EqmSizeOut) con->EqmSizeOut = k;
  return cc;
}
#define prpqs2pqs(p) {\
  strncpy(pqs[p].prpName,prpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(pqs[p].prpDescription,prpqs[p].prpDescription,32);\
  pqs[p].prpSize = (UINT16)prpqs[p].prpSize;\
  pqs[p].prpFormat = prpqs[p].prpFormat;\
  pqs[p].prpAccess = prpqs[p].prpAccess;\
}
#define prpqs2pqsx(p) {\
  strncpy(pqsx[p].prpName,prpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(pqsx[p].prpDescription,prpqs[p].prpDescription,64);\
  strncpy(pqsx[p].prpRedirection,&prpqs[p].prpRedirection[PRP_REDIR_SRV_OFFSET],16);\
  strncpy(&pqsx[p].prpRedirection[16],&prpqs[p].prpRedirection[PRP_REDIR_PRP_OFFSET],16);\
  strncpy(pqsx[p].prpTag,prpqs[p].prpTag,TAG_NAME_SHORTSIZE);\
  strncpy(pqsx[p].prpTagIn,prpqs[p].prpTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(pqsx[p].prpUnits,prpqs[p].prpUnits,16);\
  pqsx[p].prpMinValue = prpqs[p].prpMinValue;\
  pqsx[p].prpMaxValue = prpqs[p].prpMaxValue;\
  pqsx[p].prpSize = prpqs[p].prpSize;\
  pqsx[p].prpSizeIn = prpqs[p].prpSizeIn;\
  pqsx[p].prpNumOverloads = prpqs[p].prpNumOverloads;\
  pqsx[p].prpHistoryDepthShort = prpqs[p].prpHistoryDepthShort;\
  pqsx[p].prpHistoryDepthLong = prpqs[p].prpHistoryDepthLong;\
  pqsx[p].prpFormat = prpqs[p].prpFormat;\
  pqsx[p].prpFormatIn = prpqs[p].prpFormatIn;\
  pqsx[p].prpAccess = prpqs[p].prpAccess;\
  pqsx[p].prpGraphType = prpqs[p].prpGraphType;\
}
#define prpqs2xpqs(p) {\
  strncpy(xpqs[p].prpName,prpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(xpqs[p].prpDescription,prpqs[p].prpDescription,64);\
  strncpy(xpqs[p].prpRedirection,&prpqs[p].prpRedirection[PRP_REDIR_SRV_OFFSET],16);\
  strncpy(&xpqs[p].prpRedirection[16],&prpqs[p].prpRedirection[PRP_REDIR_PRP_OFFSET],16);\
  strncpy(xpqs[p].prpTag,prpqs[p].prpTag,TAG_NAME_SHORTSIZE);\
  strncpy(xpqs[p].prpTagIn,prpqs[p].prpTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(xpqs[p].prpUnits,prpqs[p].prpUnits,16);\
  xpqs[p].prpMinValue = prpqs[p].prpMinValue;\
  xpqs[p].prpMaxValue = prpqs[p].prpMaxValue;\
  xpqs[p].prpSize = prpqs[p].prpSize;\
  xpqs[p].prpSizeIn = prpqs[p].prpSizeIn;\
  xpqs[p].prpNumOverloads = prpqs[p].prpNumOverloads;\
  xpqs[p].prpHistoryDepthShort = prpqs[p].prpHistoryDepthShort;\
  xpqs[p].prpHistoryDepthLong = prpqs[p].prpHistoryDepthLong;\
  xpqs[p].prpFormat = prpqs[p].prpFormat;\
  xpqs[p].prpFormatIn = prpqs[p].prpFormatIn;\
  xpqs[p].prpAccess = prpqs[p].prpAccess;\
  xpqs[p].prpGraphType = prpqs[p].prpGraphType;\
  strncpy(xpqs[p].rngUnits,prpqs[p].rngUnits,16);\
  xpqs[p].rngMinValue = prpqs[p].rngMinValue;\
  xpqs[p].rngMaxValue = prpqs[p].rngMaxValue;\
  xpqs[p].numRows = prpqs[p].numRows;\
  xpqs[p].rowSize = prpqs[p].rowSize;\
  xpqs[p].prpArrayType = prpqs[p].prpArrayType;\
  memcpy(xpqs[p].reserved,prpqs[p].reserved,6);\
}

#define xpqs2prpqs(p) {\
  memset(&prpqs[p],0,sizeof(PrpQueryStruct));\
  strncpy(prpqs[p].prpName,xpqs[p].prpName,PROPERTY_NAME_SHORTSIZE);\
  strncpy(prpqs[p].prpDescription,xpqs[p].prpDescription,64);\
  strncpy(&prpqs[p].prpRedirection[PRP_REDIR_SRV_OFFSET],xpqs[p].prpRedirection,16);\
  strncpy(&prpqs[p].prpRedirection[PRP_REDIR_PRP_OFFSET],&xpqs[p].prpRedirection[16],16);\
  strncpy(prpqs[p].prpTag,xpqs[p].prpTag,TAG_NAME_SHORTSIZE);\
  strncpy(prpqs[p].prpTagIn,xpqs[p].prpTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(prpqs[p].prpUnits,xpqs[p].prpUnits,16);\
  prpqs[p].prpMinValue = xpqs[p].prpMinValue;\
  prpqs[p].prpMaxValue = xpqs[p].prpMaxValue;\
  prpqs[p].prpSize = xpqs[p].prpSize;\
  prpqs[p].prpSizeIn = xpqs[p].prpSizeIn;\
  prpqs[p].prpNumOverloads = xpqs[p].prpNumOverloads;\
  prpqs[p].prpHistoryDepthShort = xpqs[p].prpHistoryDepthShort;\
  prpqs[p].prpHistoryDepthLong = xpqs[p].prpHistoryDepthLong;\
  prpqs[p].prpFormat = xpqs[p].prpFormat;\
  prpqs[p].prpFormatIn = xpqs[p].prpFormatIn;\
  prpqs[p].prpAccess = xpqs[p].prpAccess;\
  prpqs[p].prpGraphType = xpqs[p].prpGraphType;\
  strncpy(prpqs[p].rngUnits,xpqs[p].rngUnits,16);\
  prpqs[p].rngMinValue = xpqs[p].rngMinValue;\
  prpqs[p].rngMaxValue = xpqs[p].rngMaxValue;\
  prpqs[p].numRows = xpqs[p].numRows;\
  prpqs[p].rowSize = xpqs[p].rowSize;\
  prpqs[p].prpArrayType = xpqs[p].prpArrayType;\
  memcpy(prpqs[p].reserved,xpqs[p].reserved,6);\
}

int stockpropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,k,n,fmt,len;
  PropertyQueryStruct *pqs = (PropertyQueryStruct *)tmpWorkArea;
  PropertyQueryStructEx *pqsx = (PropertyQueryStructEx *)tmpWorkArea;
  XPropertyQueryStruct *xpqs = (XPropertyQueryStruct *)tmpWorkArea;
  PrpQueryStruct *prpqs = (PrpQueryStruct *)tmpWorkArea;
  int tgtprplen = 0;
  char *c,tgtprp[PROPERTY_NAME_SIZE+4];

  if ((len=(int)con->EqmSizeIn) > 0)
  {
    memset(tgtprp,0,PROPERTY_NAME_SIZE+4);
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_TEXT:
        if (len < 1 || len > PROPERTY_NAME_SIZE) return illegal_read_write;
        strncpy(tgtprp,dataIn,len);
        break;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME16 *)dataIn)->name,16);
        break;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME32 *)dataIn)->name,32);
        break;
      case CF_NAME64:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME64 *)dataIn)->name,64);
        break;
      default:  return illegal_format;
    }
    if ((c=strchr(tgtprp,'*')) != NULL)
    {
      *c = 0;
      tgtprplen = (int)strlen(tgtprp);
    }
    else tgtprplen = 32;
  }
  switch (fmt=LFMT(con->EqmFormatOut))
  {
    case CF_NAME64:
    case CF_NAME48:
    case CF_NAME32:
    case CF_NAME16:
      if (con->EqmSizeOut == 0) return illegal_read_write;
      len = fmtsizeof(fmt);
      for (i=0,k=0; i<(int)con->EqmSizeOut && i<NrStockProperties; i++)
      {
        if (tgtprplen && strnicmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strnicmp(data+(k-1)*len,StockProperty[i].prpName,len)) continue;
        strncpy(data+k*len,StockProperty[i].prpName,len);
        k++;
      }
      if (k < (int)con->EqmSizeOut) con->EqmSizeOut = k;
      return 0;
    case CF_USTRING:
      if (con->EqmSizeOut == 0) return illegal_read_write;
      for (i=0,k=0; i<(int)con->EqmSizeOut && i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strncmp(((DUSTRING *)data)[k-1].str,StockProperty[i].prpName,32)) continue;
        strncpy(((DUSTRING *)data)[k].str,StockProperty[i].prpName,32);
        strncpy(&((DUSTRING *)data)[k].str[32],StockProperty[i].prpDescription,48);
        ((DUSTRING *)data)[k].ival = (UINT32)StockProperty[i].prpSize;
        ((DUSTRING *)data)[k].tm = (UINT32)StockProperty[i].prpFormat;
        k++;
      }
      if (k < (int)con->EqmSizeOut) con->EqmSizeOut = k;
      return 0;
    case CF_STRUCT:
      for (i=0,k=0; i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        memset(&prpqs[k],0,sizeof(PrpQueryStruct));
        strncpy(prpqs[k].prpName,StockProperty[i].prpName,PROPERTY_NAME_SIZE);
        strncpy(prpqs[k].prpDescription,StockProperty[i].prpDescription,64);
        prpqs[k].prpSize = StockProperty[i].prpSize;
        prpqs[k].prpSizeIn = StockProperty[i].prpSizeIn;
        prpqs[k].prpFormat = BFMT(StockProperty[i].prpFormat);
        prpqs[k].prpFormatIn = BFMT(StockProperty[i].prpFormatIn);
        prpqs[k].prpAccess = (BYTE)StockProperty[i].prpAccessMode;
        prpqs[k].prpArrayType = StockProperty[i].prpArrayType;
        strncpy(prpqs[k].prpTag,StockProperty[i].prpStructTag,TAG_NAME_SIZE);
        k++;
      }
      if (tgtprplen) for (i=0; i<k; i++) prpqs[i].prpNumOverloads = k;
      if (con->EqmSizeOut%PRPQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"PRPQSr4",TAG_NAME_SIZE))
      {
        n = (UINT16)con->EqmSizeOut/PRPQUERYSTRUCT_SIZE;
        if (k > n) k = n;
        PrepOutgoingPRPQS(prpqs,data,k);
        if (k < n) con->EqmSizeOut = k*PRPQUERYSTRUCT_SIZE;
        return 0;
      }
      if (con->EqmSizeOut%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",TAG_NAME_SIZE))
      {
        n = (UINT16)con->EqmSizeOut/XPROPERTYQUERYSTRUCT_SIZE;
        if (k > n) k = n;
        for (i=0; i<n; i++) prpqs2xpqs(i);
        PrepOutgoingXPQS(xpqs,data,k);
        if (k < n) con->EqmSizeOut = k*XPROPERTYQUERYSTRUCT_SIZE;
        return 0;
      }
      if (con->EqmSizeOut%PROPERTYQUERYSTRUCTEX_SIZE == 0)
      {
        n = (int)(con->EqmSizeOut/PROPERTYQUERYSTRUCTEX_SIZE);
        if (k > n) k = n;
        for (i=0; i<n; i++) prpqs2pqsx(i);
        PrepOutgoingPQSX(pqsx,data,k);
        if (k < n) con->EqmSizeOut = k*PROPERTYQUERYSTRUCTEX_SIZE;
        return 0;
      }
      if (con->EqmSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
      {
        n = (int)(con->EqmSizeOut/PROPERTYQUERYSTRUCT_SIZE);
        if (k > n) k = n;
        for (i=0; i<n; i++) prpqs2pqs(i);
        PrepOutgoingPQS(pqs,data,k);
        if (k < n) con->EqmSizeOut = k*PROPERTYQUERYSTRUCT_SIZE;
        return 0;
      }
      return illegal_format;
    default:
      break;
  }
  return illegal_format;
}
int nstockpropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return T_numQuery(con,data,(short)NrStockProperties);
}
int historyRecordDepth(char *prpName,UINT16 *shortDepth,UINT16 *longDepth)
{
  int i;
  for (i=0; i<nHistoryRecords && strnicmp(prpName,hstTbl[i].c.EqmProperty,PROPERTY_NAME_SIZE); i++);
  if (i < nHistoryRecords) 
  {
    if (shortDepth != NULL) *shortDepth = hstTbl[i].depthShort;
    if (longDepth != NULL) *longDepth = hstTbl[i].depthLong;
    return hstTbl[i].depthShort;
  }
  return 0;
}
int lclPrpQueryFcn(ExportListStruct *el,char *dev,PrpQueryStruct *prpqs,int num,int start,int useAliasName)
{
  ExportPropertyListStruct *prp;
  int idx,j,k,i,n,devnr=-1;
  /* char *bp=NULL,*mp=NULL; */

  if (dev != NULL) devnr = GetDeviceNumber(el->EqmName,dev);
  for (i=0,j=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && i<num; idx++)
    for (prp=el->EqmProperty[idx]; prp && i<num; prp=prp->next)
  {
    if (el->EqmDevicePrpList != NULL && devnr >=0 && el->EqmDevicePrpList[devnr].prpListSize > 0) 
    { /* there's a device-specific property list */
      for (k=0; k<el->EqmDevicePrpList[devnr].prpListSize; k++) 
        if (!strnicmp(el->EqmDevicePrpList[devnr].prpList[k].name,prp->prpName,PROPERTY_NAME_SIZE)) break;
      if (k == el->EqmDevicePrpList[devnr].prpListSize) continue; /* not found in list */
    }
    if (j++ < start) continue;
    memset(&prpqs[i],0,sizeof(PrpQueryStruct));
    strncpy(prpqs[i].prpName,prp->prpName,PROPERTY_NAME_SIZE);
    strncpy(prpqs[i].prpDescription,prp->prpDescription,PROPERTY_DESC_SIZE);
    prpqs[i].prpSize = prp->prpSize;
    prpqs[i].prpSizeIn = prp->prpSizeIn;
    prpqs[i].prpFormat = BFMT(prp->prpFormat);
    prpqs[i].prpFormatIn = BFMT(prp->prpFormatIn);
    strncpy(prpqs[i].prpTag,prp->prpFormatTag.name,TAG_NAME_SIZE);
    strncpy(prpqs[i].prpTagIn,prp->prpFormatTagIn.name,TAG_NAME_SIZE);
    prpqs[i].prpGraphType = prp->egu.graph;
    prpqs[i].prpMaxValue = (float)prp->egu.max;
    prpqs[i].prpMinValue = (float)prp->egu.min;
    if (prp->prd != NULL) memcpy(prpqs[i].prpRedirection,prp->prd->prpRedirection,STATUS_SIZE);
    strncpy(prpqs[i].prpUnits,prp->egu.units,16);
    prpqs[i].prpNumOverloads++;
    prpqs[i].prpAccess = (BYTE)prp->prpAccessMode;
    historyRecordDepth(prp->prpName,&prpqs[i].prpHistoryDepthShort,&prpqs[i].prpHistoryDepthLong);
    prpqs[i].prpArrayType = prp->prpArrayType;
    prpqs[i].numRows = prp->numRows;
    prpqs[i].rowSize = prp->rowSize;
    prpqs[i].rngMaxValue = prp->xegu.max;
    prpqs[i].rngMinValue = prp->xegu.min;
    strncpy(prpqs[i].rngUnits,prp->xegu.units,16);
    n = i++;
    if (!useAliasName && prp->prpAlias[0] != 0  && i > 0)
    {
      prpqs[i] = prpqs[n];
      strncpy(prpqs[i].prpName,prp->prpAlias,PROPERTY_NAME_SIZE);
      i++;
    }
/*  // brute force scanning takes way too much time when there's a large alias table
    // only way to provide this information at browse time is to maintain a list
    // in the ExportPropertyListStruct which is filled in when an alias is registered
    // which is a meta property of the target property.
    for (i=0; i<nAliasTableEntries; i++)
    {
      if (IsMetaProperty(gAliasTable[i].name,&bp,&mp) && !strcmp(xpqs[n].prpName,bp))
      {
        prpqs[i] = prpqs[n];
        if (strlen(prpqs[i].prpDescription) < 48)
        {
          strcat(prpqs[i].prpDescription," ");
          strncat(prpqs[i].prpDescription,mp,15);
        }
        strncpy(prpqs[i].prpName,gAliasTable[i].alias,PROPERTY_NAME_SIZE);
        i++;
      }
    }
*/
  }
  return i;
}
int getSrvAddressParams(CONTRACT *con,char *dataIn,char *data)
{
  DTYPE d;
  char *tgtcon,*tgteqm,*tgtexp;
  ExportListStruct *el = getExportListItem(con->EqmName);
  int len = 0;
  if (isServiceRequest(con))
  {
    if (con->EqmSizeIn != 3) return illegal_read_write;
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
       len = fmtsizeof(LFMT(con->EqmFormatIn));
       break;
      default:
        return illegal_format;
    }
    tgtcon = dataIn; tgteqm = dataIn+len; tgtexp = dataIn+len*2;
    if (strlen(tgteqm) == 0 && strlen(tgtexp) == 0) return invalid_data;
    if (strlen(tgtcon) && strnicmp(tgtcon,gDeviceContext,MIN(CONTEXT_NAME_SIZE,len))) return invalid_data;
    for (el=ExportList; el != NULL; el=el->next)
    {
      if (strlen(tgteqm) && strncmp(tgteqm,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
      if (strlen(tgtexp) && strnicmp(tgtexp,el->EqmExportName,MIN(EXPORT_NAME_SIZE,len))) continue;
      break; /* this looks like me ! */
    }
    if (el == NULL) return invalid_data;
  }
  if (el == NULL) return non_existent_elem;
  d.dArrayLength = con->EqmSizeOut;
  d.dFormat = LFMT(con->EqmFormatOut);
  d.dTag[0] = 0;
  d.data.cptr = data;
  return getSrvAddress(&d,el->EqmName,NULL);
}
int getLastAccessList(CONTRACT *con,char *dataIn,char *data)
{
  int slen = 0;
  dataIn = dataIn;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME16:slen = 16; break;
    case CF_NAME32:slen = 32; break;
    default: return illegal_format;
  }
  switch (con->EqmSizeOut)
  {
    case 5: strncpy(data + slen*4,getDataTimeString(gLastWriteAccessTime,-1),slen);
    case 4: strncpy(data + slen*3,gLastWriteAccessDev,slen);
    case 3: strncpy(data + slen*2,gLastWriteAccessPrp,slen);
    case 2: strncpy(data + slen*1,gLastWriteAccessAddr,ADDR_SIZE);
    case 1: strncpy(data,gLastWriteAccessUsr,USERNAME_SIZE);
      break;
    default:
      return dimension_error;
  }
  return 0;
}
int propsQuery(CONTRACT *con,char *dataIn,char *data)
{
  static BYTE *tmpDataCache = (BYTE *)NULL;
  static int tmpDataCacheSize = 0;
  int idx,j=0,k,n,num,len;
  short fmt;
  ExportListStruct *el;
  PropertyQueryStruct *pqs = (PropertyQueryStruct *)tmpDataCache;
  PropertyQueryStructEx *pqsx = (PropertyQueryStructEx *)tmpDataCache;
  XPropertyQueryStruct *xpqs = (XPropertyQueryStruct *)tmpDataCache;
  PrpQueryStruct *prpqs = (PrpQueryStruct *)tmpDataCache;
  int cc = 0,tgtprplen = 0,useAliasName=FALSE,usepqs=FALSE,usepqsx=FALSE,uselist=FALSE,hasdevgroup=FALSE;
  char tgtprp[PROPERTY_NAME_SIZE+4];
  char btok[PROPERTY_NAME_SIZE], etok[PROPERTY_NAME_SIZE], mtok[PROPERTY_NAME_SIZE];
  int blen = 0, elen = 0, mlen = 0, checkWildCard = FALSE;

  if (tmpDataCache == NULL)
  { /* on INTEL machines, tmpWorkArea points to the srvWorkArea */
    tmpDataCache = tmpWorkArea;
    tmpDataCacheSize = tmpWorkAreaSize;
  }
  if (con->EqmSizeOut == 0) return illegal_read_write;

  if (strcmp(con->EqmProperty,"*") && strchr(con->EqmProperty,'*') != NULL)
  { /* property names query with non-trivial wildcard string */
    getWildcardTokens(con->EqmProperty,btok,etok,mtok);
    blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
    checkWildCard = TRUE;
  }

  if ((len=(int)con->EqmSizeIn) > 0)
  {
    memset(tgtprp,0,PROPERTY_NAME_SIZE+4);
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_TEXT:
        if (len < 1 || len > PROPERTY_NAME_SIZE) return illegal_read_write;
        strncpy(tgtprp,dataIn,len);
        break;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME16 *)dataIn)->name,16);
        break;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME32 *)dataIn)->name,32);
        break;
      case CF_NAME64:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME64 *)dataIn)->name,64);
        break;
      default:  return illegal_format;
    }
    if (nAliasTableEntries > 0) /* is the incoming name an alias ? */
    {
      if ((j=getAliasTableIndex(tgtprp)) >= 0) 
      {
        if (!IsMetaProperty(gAliasTable[j].name,NULL,NULL))
        {
          strncpy(tgtprp,gAliasTable[j].name,PROPERTY_NAME_SIZE);
          useAliasName = TRUE;
        }
      }
    }
    tgtprplen = (int)strlen(tgtprp);
  }
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (strlen(el->EqmGroup) > 0) hasdevgroup = TRUE;
    len = tmpDataCacheSize/XPROPERTYQUERYSTRUCT_SIZE;
    if (el->EqmPropertyQueryFcn != NULL)           
    { /* registered query fcn */
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (num > 0 && tmpDataCache != (BYTE *)prpqs)
      { /* cache area not pointing to the query struct area */
        if (num > len)
        { /* not enough room -> re-allocate it */
          if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
            return out_of_server_memory;
          tmpDataCacheSize = num * sizeof(PrpQueryStruct);
          feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
        }
        memcpy(tmpDataCache,prpqs,num*sizeof(PrpQueryStruct));
        prpqs = (PrpQueryStruct *)tmpDataCache;
      }
      usepqs = TRUE;
    }
    else if (el->XEqmPropertyQueryFcn != NULL)        /* registered ext query fcn */
    {
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (num > 0 && tmpDataCache != (BYTE *)xpqs)
      {
        if (num > len)
        {
          if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
            return out_of_server_memory;
          tmpDataCacheSize = num * sizeof(PrpQueryStruct);
          feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
        }
        memcpy(tmpDataCache,xpqs,num*sizeof(XPropertyQueryStruct));
        prpqs = (PrpQueryStruct *)tmpDataCache;
      }
      for (k=0; k<num; k++) xpqs2prpqs(k);
      usepqsx = TRUE;
    }
    else                                                      /* handle it here */
    {
      if ((num=(el->EqmNumProperties*2)) > len) /* allow room for alias entries */
      {
        if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
          return out_of_server_memory;
        tmpDataCacheSize = num * sizeof(PrpQueryStruct);
        feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
      }
      prpqs = (PrpQueryStruct *)tmpDataCache;
      num = lclPrpQueryFcn(el,con->EqmDeviceName,prpqs,num,0,useAliasName);
      if (el->EqmDevicePrpList != NULL) uselist = TRUE;
    }
    if (num <= 0) return -num;
    /* re-assign these just in case they've changed */
    pqs = (PropertyQueryStruct *)tmpDataCache;
    pqsx = (PropertyQueryStructEx *)tmpDataCache;
    xpqs = (XPropertyQueryStruct *)tmpDataCache;
    prpqs = (PrpQueryStruct *)tmpDataCache;

    /* do we have to collapse the query results ? */
    if (checkWildCard)
    { /* checking a wildcard match */
      for (idx=0,k=0; idx<num && k<num; idx++)
      {
        if (!wildcardMatches(prpqs[idx].prpName,btok,blen,etok,elen,mtok,mlen))
          continue;
        if (idx != k) memcpy(&prpqs[k],&prpqs[idx],sizeof(PrpQueryStruct));
        k++;
      }
      num = k;
    }
    else if (tgtprplen > 0)
    { /* checking for input target property */
      for (idx=0,k=0; idx<num && k<num; idx++)
      {
        if (strnicmp(prpqs[idx].prpName,tgtprp,PROPERTY_NAME_SIZE)) continue;
        if (idx != k) memcpy(&prpqs[k],&prpqs[idx],sizeof(PrpQueryStruct));
        if (useAliasName) strncpy(prpqs[k].prpName,gAliasTable[j].alias,PROPERTY_NAME_SIZE);
        k++;
      }
      for (j=0; j<k; j++) prpqs[j].prpNumOverloads = k;
      num = k;
    }
    switch (fmt=LFMT(con->EqmFormatOut))
    {
      case CF_STRUCT:
        if (num == 0) return non_existent;
        if (con->EqmSizeOut%PRPQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"PRPQSr4",TAG_NAME_SIZE))
        {
          n = (int)con->EqmSizeOut/PRPQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          PrepOutgoingPRPQS(prpqs,data,num);
          if (num < n) con->EqmSizeOut = num*PRPQUERYSTRUCT_SIZE;
          return 0;
        }
        if (con->EqmSizeOut%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",TAG_NAME_SIZE))
        {
          n = (int)con->EqmSizeOut/XPROPERTYQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          for (k=0; k<num; k++) prpqs2xpqs(k);
          PrepOutgoingXPQS(xpqs,data,num);
          if (num < n) con->EqmSizeOut = num*XPROPERTYQUERYSTRUCT_SIZE;
          return 0;
        }
        if (con->EqmSizeOut%PROPERTYQUERYSTRUCTEX_SIZE == 0)
        {
          n = (int)con->EqmSizeOut/PROPERTYQUERYSTRUCTEX_SIZE;
          if (num > n) num = n;
          for (k=0; k<num; k++) prpqs2pqsx(k);
          PrepOutgoingPQSX(pqsx,data,num);
          if (num < n) con->EqmSizeOut = num*PROPERTYQUERYSTRUCTEX_SIZE;
          return 0;
        }
        if (con->EqmSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
        {
          n = (int)con->EqmSizeOut/PROPERTYQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          if (usepqs && (checkWildCard || tgtprplen > 0))
          {
            pqs = (PropertyQueryStruct *)tmpDataCache;
            usepqs = FALSE;
          }
          if (!usepqs) for (k=0; k<num; k++) prpqs2pqs(k);
          PrepOutgoingPQS(pqs,data,num);
          if (num < n) con->EqmSizeOut = num*PROPERTYQUERYSTRUCT_SIZE;
          return 0;
        }
        return illegal_format;
      case CF_NAME64:
      case CF_NAME48:
      case CF_NAME32:
      case CF_NAME16:
        n = (int)con->EqmSizeOut;
        if (num > n) num = n;
        len = fmtsizeof(fmt);
        for (idx=0,k=0; idx<num && k<n; idx++)
        { /* check for up to 2 overloads */
          if (k > 0 && !strncmp(data+(k-1)*len,prpqs[idx].prpName,len)) continue;
          if (k > 1 && !strncmp(data+(k-2)*len,prpqs[idx].prpName,len)) continue;
          strncpy(data+k*len,prpqs[idx].prpName,len);
          k++;
        }
        if (k == 0) return non_existent;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        if (k < n) con->EqmSizeOut = k;
        return cc;
      case CF_USTRING:
        n = (int)con->EqmSizeOut;
        if (num > n) num = n;
        for (idx=0,k=0; idx<num && k<n; idx++)
        { /* check for up to 2 overloads */
          if (k > 0 && !strncmp(((DUSTRING *)data)[k-1].str,prpqs[idx].prpName,32)) continue;
          if (k > 1 && !strncmp(((DUSTRING *)data)[k-2].str,prpqs[idx].prpName,32)) continue;
          strncpy(((DUSTRING *)data)[k].str,prpqs[idx].prpName,32);
          strncpy(&((DUSTRING *)data)[k].str[32],prpqs[idx].prpDescription,48);
          ((DUSTRING *)data)[k].ival = (UINT32)prpqs[idx].prpSize;
          ((DUSTRING *)data)[k].tm = (UINT32)prpqs[idx].prpFormat;
          k++;
        }
        if (k == 0) return non_existent;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        if (k < n) con->EqmSizeOut = k;
        return cc;
      default:
        return illegal_format;
    }
  }
  return non_existent_elem;
}
int npropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el;
  dataIn = dataIn;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    return T_numQuery(con,data,el->EqmNumProperties);
  }
  return non_existent_elem;
}
int alarmsQuery(CONTRACT *con,char *dataIn,char *data)
{
  AMS *ams;
  AMRB amrb;
  int n,nalm,minsev,start,isLegacyCall=FALSE;
  int size = (int)con->EqmSizeOut;
  int sizeIn = (int)con->EqmSizeIn;
  UINT32 startt, stopt;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (sizeIn > 0)
  { /* Note: allow input to be an array of UINT32s as well !! */
    if (LFMT(con->EqmFormatIn) != CF_STRUCT) return illegal_format;
    if (sizeIn != AMRB_SIZE) return dimension_error;
    prepIncomingAMRB(&amrb,(BYTE *)dataIn);
    stopt = amrb.almStopTime;
    startt = amrb.almStartTime;
    minsev = amrb.almMinSeverity;
  }
  else
  {
    stopt = 0x7fffffffL;
    startt = 0;
    minsev = 0;
  }
  if (strcmp(con->strTagOut,"AMSr4") != 0)
  { /* old-style request (could be untagged) */
    if (con->EqmSizeOut%AMS5_SIZE) return illegal_read_write;
    n = (int)con->EqmSizeOut/AMS5_SIZE;
    isLegacyCall = TRUE;
  }
  else
  {
    n = (int)con->EqmSizeOut/AMS_SIZE;
  }
  ams = (AMS *)tmpWorkArea;
  start = con->EqmDeviceName[0] == '#' ? atoi(&con->EqmDeviceName[1]) : 0;
  nalm = getAlarmsEx(con->EqmName,ams,(short)n,(short)minsev,(time_t)startt,(time_t)stopt,(short)start);
  prepOutgoingAMS((BYTE *)data,ams,con->strTagOut,nalm);  
  if (nalm < n) con->EqmSizeOut = nalm * (isLegacyCall ? AMS5_SIZE : AMS_SIZE);
  return 0;
}
int nalarmsQuery(CONTRACT *con,char *dataIn,char *data)
{
  short sc,ntim,nsev;

  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_LONG) return illegal_format;
  switch (con->EqmSizeOut)
  {
    case 3:
      *(UINT32 *)data = numAlarmsEx(con->EqmName,&((UINT32 *)data)[1],&sc,&ntim,&nsev);
      ((UINT32 *)data)[2] = (UINT32)sc;
      break;
    case 6:
      ((UINT32 *)data)[5] = numAlarmDefs(con->EqmName);
    case 5:
      *(UINT32 *)data = numAlarmsEx(con->EqmName,&((UINT32 *)data)[1],&sc,&ntim,&nsev);
      ((UINT32 *)data)[2] = (UINT32)sc;
      ((UINT32 *)data)[3] = (UINT32)ntim;
      ((UINT32 *)data)[4] = (UINT32)nsev;
      break;
    default: 
      return illegal_read_write;
  }
  if (!gRequireReadFromCAS) 
  { /* CAS has persistent link  to NALARMS ! */
    if ((gRequireReadFromCAS=isCASInUserList(con->EqmName)) != FALSE) 
      feclog("ALM: Central Alarm Server attached");
  }
  return 0;
}
int nalarmWatchQuery(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return T_numQuery(con,data,(short)nAlmWatchList);
}
int alarmWatchQuery(CONTRACT *con,char *dataIn,char *data)
{
  AWSList *lst = NULL;
  AWS aws;
  int cc, n;
  int sizeIn = (int)con->EqmSizeIn;
  ExportListStruct *el;
  ExportPropertyListStruct *pl;
  char sts[STATUS_SIZE];

  if (con->EqmAccess&CA_WRITE)
  {
    if (LFMT(con->EqmFormatIn) != CF_STRUCT) return illegal_format;
    if ((cc=prepIncomingAWS((BYTE *)dataIn,&aws,con->strTagIn,1)) != 0) return cc;
    /* enforce the local eqm name : */
    strncpy(aws.eqm,con->EqmName,EQM_NAME_SIZE);
    for (lst=almWatchList; lst != NULL; lst=lst->nxt)
    { /* find the entry */
      if (strncmp(lst->aws.eqm,aws.eqm,EQM_NAME_SIZE)) continue;
      if (strnicmp(lst->aws.dev,aws.dev,DEVICE_NAME_SIZE)) continue;
      if (strnicmp(lst->aws.prp,aws.prp,PROPERTY_NAME_SIZE)) continue;
      /* it's in the current list, so enforce the following */
      aws.siz = lst->aws.siz;
      aws.fmt = lst->aws.fmt;
      aws.atyp = lst->aws.atyp;
      break;
    }
    if (lst == NULL)
    { /* was not in the current list */
      if ((cc=IsIllegalProperty(aws.eqm,aws.prp,aws.dev,NULL,CA_READ,&el,sts)) != 0) return cc;
      if ((lst=(AWSList *)SystemCalloc(1,sizeof(AWSList))) == NULL) return out_of_server_memory;
      if ((pl=GetProperyListStruct(aws.eqm,aws.prp)) != NULL)
      { /* does not use a property query function (good news!) */
        if (aws.siz == 0 || aws.siz > (int)pl->prpSize) aws.siz = pl->prpSize;
        aws.fmt = pl->prpFormat;
        aws.atyp = pl->prpArrayType;
      }
      /* add it to the list */
      memcpy(&lst->aws,&aws,sizeof(AWS));
      lst->nxt = almWatchList;
      almWatchList->nxt = lst;
      nAlmWatchList++;
    }
    else
    { /* apply the changes */
      memcpy(&lst->aws,&aws,sizeof(AWS));
    }
  }
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
    n = (int)con->EqmSizeOut/AWS_SIZE;
    if (nAlmWatchList > 0)
    {
      if ((cc=prepOutgoingAWS((BYTE *)data,almWatchList,con->strTagOut,n)) != 0) return cc;
    }
    if (nAlmWatchList < n) con->EqmSizeOut = nAlmWatchList * AWS_SIZE;
  }
  return 0;
}
int alarmDefsQuery(CONTRACT *con,char *dataIn,char *data)
{ 
  ADSList *adslst = NULL;
  int n, isLegacyCall=FALSE;
  short num = 0;
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;

  if ((num=getAlarmDefs(con->EqmName,&adslst)) < 0) return -num;
  if (num == 0) return un_allocated;
  if (strcmp(con->strTagOut,"ADSr4") != 0)
  { /* old-style request (could be untagged) */
    if (con->EqmSizeOut%ADS5_SIZE) return illegal_read_write;
    n = (int)con->EqmSizeOut/ADS5_SIZE;
    isLegacyCall = TRUE;
  }
  else
  {
    n = (int)con->EqmSizeOut/ADS_SIZE;
  }
  prepOutgoingADS((BYTE *)data,adslst,con->strTagOut,n);

  if (num < n) con->EqmSizeOut = num * (isLegacyCall ? ADS5_SIZE : ADS_SIZE);

  return 0;
}
int nalarmDefsQuery(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return T_numQuery(con,data,(short)numAlarmDefs(con->EqmName));
}
int msgDisplay(CONTRACT *con,char *dataIn,char *data)
{
  int len = (int)con->EqmSizeIn;
  data = data;
  if (len == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatIn) != CF_TEXT) return illegal_format;
  memset(dbgbuf,0,256);
  strncpy(dbgbuf,dataIn,MIN(len,255));
  dbglog(dbgbuf);
  feclog(dbgbuf);
  msglog(0,dbgbuf);
  return 0;
}
int activityQuery(CONTRACT *con,char *dataIn,char *data)
{
  ActivityQueryStruct aqs;

  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqmSizeOut != ACTIVITYQUERYSTRUCT_SIZE) return illegal_read_write;

  memset(&aqs.reserved,0,4);
  strncpy(aqs.FecName,gFecName,FEC_NAME_SIZE);
  aqs.localtime = (UINT32)time(NULL);
  aqs.starttime = (UINT32)startuptime;
  aqs.systemPollingRate = SysPoll;
  aqs.numBkgTsks = gNumEqmBkgTasks;
  aqs.numTotalContracts = ncontracts - nExpiredContracts;
  aqs.numTotalClients = nconsumers;
  aqs.numTargetContracts = GetNumContracts(con->EqmName);
  aqs.numTargetClients = GetNumConsumers(con->EqmName);
  aqs.numConnections = nConnectionTableEntries;
  aqs.numConnectionTimeouts = nConnectionTimeouts;
  aqs.numConnectionArrivals = nArrivals;
  aqs.numUDPpkts = UDPcount;
  aqs.numTCPpkts = TCPcount;
  aqs.numIPXpkts = IPXcount;
  aqs.numSPXpkts = SPXcount;

  PrepOutgoingAQS(&aqs,data);
  return 0;
}
int contractsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ContractQueryStruct *cqs = NULL;
  Contract5QueryStruct *cqs5 = NULL;
  ContractListStruct *cl;
  int i, n, ncon;
# if !defined(NONALIGNED) && !defined(INTEL)
  int tmp;
# endif
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (!strcmp(con->strTagOut,"CTQSr4") || con->EqmSizeOut%CONTRACT_QUERY_SIZE == 0)
  {
    if ((ncon=(int)con->EqmSizeOut/CONTRACT_QUERY_SIZE) == 0) return dimension_error;
    cqs = (ContractQueryStruct *)data;
  }
  if (!strcmp(con->strTagOut,"CONQS") || con->EqmSizeOut%CONTRACT5_QUERY_SIZE == 0)
  {
    if ((ncon=(int)con->EqmSizeOut/CONTRACT5_QUERY_SIZE) == 0) return dimension_error;
    cqs5 = (Contract5QueryStruct *)data;
  }
  for (i=0,n=0; n<ncon && i<ncontracts; i++)
  {
    cl = ContractList[i];
    if (cl->expired) continue;
    if (cqs != NULL)
    {
      memcpy(&cqs[n].contract,&cl->contract,sizeof(CONTRACT));
      cqs[n].pollrate = cl->PollingRate;
      cqs[n].nconsumers = cl->nconsumers;
      PrepOutgoingCONQS(cqs[n]);
    }
    else if (cqs5 != NULL)
    {
      memcpy(cqs5[n].contract.EqmProperty,cl->contract.EqmProperty,PROPERTY_NAME_SHORTSIZE);
      memcpy(cqs5[n].contract.EqmDeviceName,cl->contract.EqmDeviceName,DEVICE_NAME_SHORTSIZE);
      memcpy(cqs5[n].contract.EqmName,cl->contract.EqmName,EQM_NAME_SIZE);
      cqs5[n].contract.EqmSizeIn = cl->contract.EqmSizeIn;
      cqs5[n].contract.EqmSizeOut = cl->contract.EqmSizeOut;
      memcpy(&cqs5[n].contract.hEqmName,&cl->contract.hEqmName,4);
      memcpy(cqs5[n].contract.strTagIn,cl->contract.strTagIn,TAG_NAME_SHORTSIZE);
      memcpy(cqs5[n].contract.strTagOut,cl->contract.strTagOut,TAG_NAME_SHORTSIZE);
      cqs5[n].pollrate = cl->PollingRate;
      cqs5[n].nconsumers = cl->nconsumers;
      PrepOutgoingCONQS(cqs5[n]);
    }
    n++;
  }
  con->EqmSizeOut = n * (cqs == NULL ? CONTRACT5_QUERY_SIZE : CONTRACT_QUERY_SIZE);
  return 0;
}
int rmvRegisteredUser(CONTRACT *con,char *dataIn,char *data)
{
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqmSizeIn, lstsiz = 0;
  data = data;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqmAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,USERNAME_SIZE);
      break;
    case CF_TEXT:
      memset(usr.name,0,USERNAME_SIZE);
      if (len < 1) return illegal_read_write;
      if (len > USERNAME_SIZE) len = USERNAME_SIZE;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    default: return illegal_format;
  }
  return removeRegisteredUser(usrlst,lstsiz);
}
int addRegisteredUser(CONTRACT *con,char *dataIn,char *data)
{
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqmSizeIn, lstsiz = 0;
  data = data;
  /* If there's no list now, don't let someone make one over the net (dangerous!) */
  if (NrOfUsers == 0) return not_allowed; 
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqmAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,USERNAME_SIZE);
      break;
    case CF_TEXT:
      memset(usr.name,0,USERNAME_SIZE);
      if (len < 1) return illegal_read_write;
      if (len > USERNAME_SIZE) len = USERNAME_SIZE;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    case CF_NULL:
      /* force a re-read of users.csv */
      break;
    default: 
      return illegal_format;
  }
  return appendRegisteredUserList(usrlst,lstsiz);
}
int addRegisteredIPNet(CONTRACT *con,char *dataIn,char *data)
{
  char ip[32];
  int len = (int)con->EqmSizeIn;
  data = data;
  /* If there's no list now, don't let someone make one over the net (dangerous!) */
  if (NrOfIPNets == 0) return not_allowed; 
  memset(ip,0,17);
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
      if (len != 1) return illegal_read_write;
      strncpy(ip,((NAME16 *)dataIn)->name,16);
      break;
    case CF_NAME32:
      if (len != 1) return illegal_read_write;
      strncpy(ip,((NAME32 *)dataIn)->name,32);
      break;
    case CF_TEXT:
      if (len < 1 || len > 16) return illegal_read_write;
      strncpy(ip,dataIn,len);
      break;
    case CF_NULL:
      /* force a file re-read */
      break;
    default: 
      return illegal_format;
  }
  return appendRegisteredIPNetsList(ip);
}
int getNumRegisteredIPNets(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return T_numQuery(con,data,(short)NrOfIPNets);
}
int getRegisteredIPNetList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  char ip[32];
  dataIn = dataIn;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqmSizeOut && i<NrOfIPNets; i++)
  {
    inet_ntoa_b(IPNetList[i].sin_addr,ip);
    strncpy(((NAME16 *)data)[i].name,ip,16);
  }
  return 0;
}
int getNumRegisteredIPXNets(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
# if defined(IPXNW) || defined(IPXBSD)
  return T_numQuery(con,data,NrOfIPXNets);
# else
  data = data;
  con = con;
  return ipx_not_supported;
# endif
}
int getRegisteredIPXNetList(CONTRACT *con,char *dataIn,char *data)
{
# if defined(IPXNW) || defined(IPXBSD)
  int i,j;
  char ipx[32];
  dataIn = dataIn;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<con->EqmSizeOut && i<NrOfIPXNets; i++)
  {
    for (j=0;j<4;j++) sprintf(&ipx[j*2],"%02X",IPXNetList[i].net[j]);
    strncpy(((NAME16 *)data)[i].name,ipx,16);
  }
  return 0;
# else
  dataIn = dataIn;
  data = data;
  con = con;
  return ipx_not_supported;
# endif
}
int getNumRegisteredUsers(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return T_numQuery(con,data,(short)NrOfUsers);
}
int getFecLog(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    readFecLog(data,(int)con->EqmSizeOut);
  }
  return 0;
}
int getFecLogFile(CONTRACT *con,char *dataIn,char *data)
{
  char fn[128];
  int slen, cc;

  if (con->EqmDeviceName[0] != '#') 
  {
    strncpy(fn,con->EqmDeviceName,DEVICE_NAME_SIZE); fn[DEVICE_NAME_SIZE] = 0;
  }
  if (con->EqmSizeIn > 0)
  {
    if (con->EqmAccess & CA_WRITE)
    { /* write the text file submitted to disk */
      if (con->EqmSizeIn == 0) return illegal_read_write;
      if ((cc=writeFecLogFile(fn,dataIn,(int)con->EqmSizeIn)) != 0) return cc;
    }
    else
    { /* alternate way of supplying a (long) file name */
      switch (LFMT(con->EqmFormatIn))
      {
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          slen = fmtsizeof(con->EqmFormatIn);
          break;
        case CF_TEXT:
          slen = con->EqmSizeIn;
          break;
        default: return illegal_format;
      }
      strncpy(fn,dataIn,slen); fn[slen] = 0;
    }
  }
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    if ((cc=readFecLogFile(fn,data,(int)con->EqmSizeOut)) != 0) return cc;
    con->EqmSizeOut = (UINT32)strlen(data);
  }
  return 0;
}
int getFecFile(CONTRACT *con,char *dataIn,char *data)
{
  char fn[128];
  int slen, cc, sz;

  if (con->EqmDeviceName[0] == '#') return invalid_name;
  strncpy(fn,con->EqmDeviceName,DEVICE_NAME_SIZE); fn[DEVICE_NAME_SIZE] = 0;
  if (con->EqmSizeIn > 0)
  {
    if (con->EqmAccess & CA_WRITE)
    { /* write the text file submitted to disk */
      if (con->EqmSizeIn == 0) return illegal_read_write;
      if ((cc=writeFecFile(fn,dataIn,(int)con->EqmSizeIn)) != 0) return cc;
    }
    else
    { /* alternate way of supplying a (long) file name */
      switch (LFMT(con->EqmFormatIn))
      {
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          slen = fmtsizeof(con->EqmFormatIn);
          break;
        case CF_TEXT:
          slen = con->EqmSizeIn;
          break;
        default: return illegal_format;
      }
      strncpy(fn,dataIn,slen); fn[slen] = 0;
    }
  }
  if (con->EqmSizeOut > 0)
  {
    if (fmtsizeof(con->EqmFormatOut) != 1) return illegal_format;
    sz = (int)con->EqmSizeOut;
    if ((cc=readFecFile(fn,data,&sz)) != 0) return cc;
    con->EqmSizeOut = (UINT32)sz;
  }
  return 0;
}
int getFecLogFileList(CONTRACT *con,char *dataIn,char *data)
{
  int cc, n, slen = 0;
  char path[256];
  if (con->EqmSizeOut > 0)
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        slen = fmtsizeof(con->EqmFormatIn);
        break;
      case CF_TEXT:
        slen = con->EqmSizeIn;
        break;
      default: 
        break;
    }
    if (slen > 0) strncpy(path,dataIn,slen); 
    path[slen] = 0; 
    applyFecLogPath(path,128);
    if (LFMT(con->EqmFormatOut) != CF_NAME64) return illegal_format;
    n = (int)con->EqmSizeOut;
    if ((cc=scanFecLogFiles(path,NULL,(NAME64 *)data,&n,0)) != 0) return cc;
    con->EqmSizeOut = (UINT32)n;
  }
  return 0;
}
int getSelfTestFile(CONTRACT *con,char *dataIn,char *data)
{
  int bytesread;
  dataIn = dataIn;
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    bytesread = readSelfTest(con->EqmName,data,(int)con->EqmSizeOut);
    if (bytesread == 0) return no_such_file;
    con->EqmSizeOut = bytesread;
  }
  return 0;
}
int setLogDepth(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqmSizeIn > 0 && !nofeclog)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_SHORT) return illegal_format;
    FeclogDepth = *(short *)dataIn;
  }
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)FeclogDepth;
  }
  return 0;
}
int setLogCommands(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_SHORT) return illegal_format;
    putCommandsInFeclog = *(short *)dataIn;
  }
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)putCommandsInFeclog;
  }
  return 0;
}
int setDebugLevel(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_SHORT) return illegal_format;
    tineDebug = *(short *)dataIn;
  }
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = tineDebug;
  }
  return 0;
}
int getAppVersion(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqmSizeOut) < 1) return illegal_read_write;
  strncpy(data,appmkversion,len);
  return 0;
}
int getTINEVersion(CONTRACT *con,char *dataIn,char *data)
{
  char s[32];
  int len;
  BYTE *ver = SystemVersion();
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  sprintf(s,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  strncpy(data,s,len);
  return 0;
}
int getServerOS(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  strncpy(data,OS_TAG,len);
  return 0;
}
int getServerTimeStr(CONTRACT *con,char *data,time_t t)
{
  int len;
  char s[32];
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;

  switch (LFMT(con->EqmFormatOut))
  {
    case CF_TEXT:
      strncpy(s,ctime(&t),24); s[24] = 0;
      switch (len)
      {
        case 32: case 26: case 24: strncpy(data,s,24); break;
        case 16: strncpy(data,&s[4],15); data[15] = 0; break;
        default: return illegal_read_write;
      }
      break;
    case CF_LONG:
      *(UINT32 *)data = (UINT32)t;
      break;
    default:
      return illegal_format;
  }
  return 0;
}
int getServerStarttime(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return getServerTimeStr(con,data,startuptime);
}
int getAppDate(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  return getServerTimeStr(con,data,appmktime);
}
int getServerDescription(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  strncpy(data,gFecInfo.desc,len);
  return 0;
}
int getServerLocation(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  strncpy(data,gFecInfo.loc,len);
  return 0;
}
int getServerStats(CONTRACT *con,char *dataIn,char *data)
{
  UINT32 statsBuffer[16];
  int i;
  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_LONG) return illegal_format;
  statsBuffer[0] = AveBusyTime;
  statsBuffer[1] = gCycleCounts;
  statsBuffer[2] = gMaxCycleCounts;
  statsBuffer[3] = gSingleLinkCount;
  statsBuffer[4] = gClientMisses;
  statsBuffer[5] = gClientReconnects;
  statsBuffer[6] = gClientRetries;
  statsBuffer[7] = gContractMisses;
  statsBuffer[8] = gContractDelays;
  statsBuffer[9] = gBurstLimitReachedCount;
  statsBuffer[10] = (UINT32)(gDataTimeStampOffset * 1000);
  for (i=0; i<(int)con->EqmSizeOut && i<11; i++) ((UINT32 *)data)[i] = statsBuffer[i];
  return 0;
}
int getDeniedUserList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  dataIn = dataIn;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqmSizeOut && i<NrOfDeniedUsers; i++)
  {
    strncpy(((NAME16 *)data)[i].name,gDeniedUsersList[i].name,USERNAME_SIZE);
  }
  return 0;
}
int getRegisteredUserList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  dataIn = dataIn;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqmSizeOut && i<NrOfUsers; i++)
  {
    strncpy(((NAME16 *)data)[i].name,gAllowedUsersList[i].name,USERNAME_SIZE);
  }
  return 0;
}
int clientsQuery(CONTRACT *con,char *dataIn,char *data)
{
  CLN *cln;
  int i, ncln;

  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqmSizeOut%CLIENT_QUERY_SIZE) return illegal_read_write;
  if ((ncln=(int)con->EqmSizeOut/CLIENT_QUERY_SIZE) == 0) return dimension_error;

  for (i=0; i<ncln && i<nconsumers; i++)
  {
    if (Consumers[i] == NULL) continue;
    cln = Consumers[i];
    PrepOutgoingCLNQS(&data[i*CLIENT_QUERY_SIZE],cln);
  }
  con->EqmSizeOut = i*CLIENT_QUERY_SIZE;
  return 0;
}

int gateFunction(CONTRACT *con,char *dataIn,char *data)
{ /* keep up apperances ... */
  dataIn = dataIn;
  data = data;
  con = con;
  return 0;
}
int EGUQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int idx,num,len;
  ExportPropertyListStruct *prp,prpTmp;
  ExportListStruct *el;
  PrpQueryStruct *prpqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  char *c, Property[PROPERTY_NAME_SIZE+4];
  DUSTRING *ustr;

  dataIn = dataIn;
  if ((len=(int)con->EqmSizeOut) == 0) return illegal_read_write;
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".EGU")) == NULL) return illegal_property;
  *c = 0;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      prp = NULL;
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        /* Is .EGU query managed by query function ? */
        if (!strnicmp(con->EqmProperty,xpqs[idx].prpName,PROPERTY_NAME_SHORTSIZE)) return ambiguous;
        if (strnicmp(Property,xpqs[idx].prpName,PROPERTY_NAME_SHORTSIZE)) continue;
        strncpy(prpTmp.egu.units,xpqs[idx].prpUnits,32);
        prpTmp.egu.min = xpqs[idx].prpMinValue;
        prpTmp.egu.max = xpqs[idx].prpMaxValue;
        prpTmp.egu.graph = xpqs[idx].prpGraphType;
        prp = &prpTmp;
      }
      if (prp != NULL) goto checkformat;
      return illegal_property;
    }
    if (el->EqmPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      prp = NULL;
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        /* Is .EGU query managed by query function ? */
        if (!strnicmp(con->EqmProperty,prpqs[idx].prpName,PROPERTY_NAME_SIZE)) return ambiguous;
        if (strnicmp(Property,prpqs[idx].prpName,PROPERTY_NAME_SIZE)) continue;
        strncpy(prpTmp.egu.units,prpqs[idx].prpUnits,32);
        prpTmp.egu.min = prpqs[idx].prpMinValue;
        prpTmp.egu.max = prpqs[idx].prpMaxValue;
        prpTmp.egu.graph = prpqs[idx].prpGraphType;
        prp = &prpTmp;
      }
      if (prp != NULL) goto checkformat;
      return illegal_property;
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
    {
      if (strnicmp(Property,prp->prpName,PROPERTY_NAME_SIZE)) continue;
      goto checkformat;
    }
    if (prp == NULL)
    {
      memset(&prpTmp,0,sizeof(ExportPropertyListStruct));
      prp = &prpTmp;
    }
checkformat:
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_USTRING:
        ustr = (DUSTRING *)data;
        ustr->ival  = (UINT32)prp->egu.graph;
        ustr->f1val = (float)prp->egu.min;
        ustr->f2val = (float)prp->egu.max;
        ustr->tm = (UINT32)startuptime;
        strncpy(ustr->str,prp->egu.units,16);
        return 0;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,16);
        return 0;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,32);
        return 0;
      case CF_TEXT:
        if (len < 1 || len > 16) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,len);
        return 0;
      case CF_FLOAT:
        if (len != 2) return illegal_read_write;
        ((float *)data)[0] = (float)prp->egu.min;
        ((float *)data)[1] = (float)prp->egu.max;
        return 0;
      default:
        return illegal_format;
    }
  }
  return non_existent_elem;
}

TINE_EXPORT int gAllowRemoteExit = FALSE;
int srvExit(CONTRACT *con,char *dataIn,char *data)
{
  dataIn = dataIn;
  data = data;
  if (!gAllowRemoteExit) return not_accepted;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  feclog("remote exit from %s",GetCaller(STKEQM_NAME));
  quitServer();
  return 0;
}
TINE_EXPORT int gAllowRemoteReset = TRUE;
int srvReset(CONTRACT *con,char *dataIn,char *data)
{
  int level = 0;
  data = data;
  if (!gAllowRemoteReset) return not_accepted;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (con->EqmSizeIn > 0) 
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_SHORT:
        level = (int)((SINT16 *)dataIn)[0];
        break;
      case CF_LONG:
        level = (int)((SINT32 *)dataIn)[0];
        break;
      default:
        return illegal_format;
    }
  }
  feclog("remote reset from %s",GetCaller(STKEQM_NAME));
  SystemReset(level);
  return 0;
}
TINE_EXPORT int gAllowRemoteInit = FALSE;
int srvInit(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el = getExportListItem(con->EqmName);
  dataIn = dataIn;
  data = data;
  if (!gAllowRemoteInit) return not_accepted;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (el == NULL) return non_existent_elem;
  feclog("remote init from %s",GetCaller(STKEQM_NAME));
  el->eqmIni();
  return 0;
}

int lockAccess(CONTRACT *con,char *dataIn,char *data)
{
  short sval[2];
  int actionTaken = FALSE;
  int passed = FALSE;
  CLN *cln = gStockRecourceSet.StkCls->consumer->client;;

  if (con->EqmAccess&CA_WRITE)
  {
    if (con->EqmSizeIn != 2) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_SHORT) return illegal_format;
    sval[0] = ((short *)dataIn)[0];
    sval[1] = ((short *)dataIn)[1];
    if (!putCommandsInFeclog)
    {
      feclog("ACCESSLOCK %d called from %s",sval[0],GetCaller(STKEQM_NAME));
    }
    if (sval[0] < 0 || sval[0] > LOCK_ABORT) return out_of_range;
    if (IsMemberLockSet(cln)) passed = TRUE;
    else if (currentAccessLock.lockType != LOCK_PERSISTENT) passed = TRUE;
    else if (sval[0] == LOCK_ABORT) passed = TRUE;
    if (!passed) return access_denied;
    if (sval[0] == LOCK_ABORT || sval[0] == LOCK_UNLOCKED)
    {
      memset(&currentAccessLock,0,sizeof(currentAccessLock));
    }
    else
    {
      memcpy(&currentAccessLock.cln,cln,sizeof(CLN));
      currentAccessLock.lockType = sval[0];
      currentAccessLock.lockDuration = sval[1];
      currentAccessLock.lockStart = (UINT32)time(NULL);
    }
    actionTaken = TRUE;
  } 
  if (con->EqmSizeOut == 2)
  { 
    if (LFMT(con->EqmFormatOut) != CF_NAME32) return illegal_format;
    strncpy(data,currentAccessLock.cln.userName,USERNAME_SIZE);
    ntoa(&currentAccessLock.cln,&data[32]);
    actionTaken = TRUE;
  }

  return actionTaken ? 0 : illegal_read_write;
}
int nhistoryRecordsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,n=0;
  dataIn = dataIn;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strcmp(con->EqmName,hstTbl[i].c.EqmName)) continue;
    n++;
  }
  return T_numQuery(con,data,(short)n);
}
int historyRecordsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,n=0;

  dataIn = dataIn;
  if (LFMT(con->EqmFormatOut) != CF_NAME32) return illegal_format;
  for (i=0; i<nHistoryRecords && n<(int)con->EqmSizeOut; i++)
  {
    if (strcmp(con->EqmName,hstTbl[i].c.EqmName)) continue;
    strncpy(((NAME32 *)data)[n++].name,hstTbl[i].c.EqmProperty,32);
  }
  return 0;
}
int historyRecordsAppend(CONTRACT *con,char *dataIn,char *data)
{
#ifndef SMALL_TINE_LIB
  HRS hrs;
  HistorySpecification hspec;
  int cc,n,idx;

  data = data;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqmSizeOut%HRS_SIZE) return illegal_read_write;
  n = (UINT16)con->EqmSizeOut/HRS_SIZE;
  if (n != 1) return dimension_error;
  memset(&hrs,0,HRS_SIZE);
  PrepIncomingHRS(&hrs,dataIn);

  idx = findAvailableHistoryIndex();
  hspec.archiveRate = hrs.archiveRate;
  hspec.aTolerance = hrs.aTolerance;
  hspec.depthLong = hrs.depthLong;
  hspec.depthShort = hrs.depthShort;
  hspec.heartbeat = hrs.heartbeat;
  hspec.pollingRate = hrs.pollingRate;
  hspec.pTolerance = hrs.pTolerance;
  hspec.rhsPropertyName = NULL;
  hspec.rhsServerName = NULL;

  cc = AppendHistoryInformation(con->EqmName,hrs.property,hrs.device,hrs.datasiz,hrs.datafmt,idx,&hspec);
  
  return cc;
# else
  con = con;
  dataIn = dataIn;
  data = data;
  return not_implemented;
# endif
}
int NAMQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int cc, i,slen;
  ExportPropertyListStruct *prp;
  char *c,tgtprp[PROPERTY_NAME_SIZE+4];
  dataIn = dataIn;
  strncpy(tgtprp,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(tgtprp,".NAM")) != NULL) *c = 0;
  if ((prp=AssignDeviceListFromProperty(con->EqmName,tgtprp)) != NULL)
  {
    if (con->EqmSizeOut == 0) return illegal_read_write;
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        slen = fmtsizeof(con->EqmFormatOut); 
        break;
      default: 
        return illegal_format;
    }
    for (i=0; i<(int)con->EqmSizeOut && i<prp->numDevices; i++)
    {
      strncpy(&data[i*slen],prp->devNames[i].name,slen);
    }
    con->EqmSizeOut = i;
    return 0;
  }
  cc = deviceQuery(con,dataIn,data);
  if (cc & CE_SENDDATA) cc = 0;
  return cc;
}
int DESCQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int idx,len,num;
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  XPropertyQueryStruct *xpqs = NULL;
  PrpQueryStruct *prpqs = NULL;
  char *c, Property[PROPERTY_NAME_SIZE+4];

  dataIn = dataIn;
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".DESC")) == NULL &&
      (c=strstr(Property,".DSC")) == NULL) return illegal_property;
  *c = 0;
  len = (int)con->EqmSizeOut;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_TEXT:
      if (len < 1 || len > 64) return illegal_read_write;
      break;
    case CF_NAME16:
      if (len != 1) return illegal_read_write;
      len = 16;
      break;
    case CF_NAME32:
      if (len != 1) return illegal_read_write;
      len = 32;
      break;
    case CF_NAME48:
      if (len != 1) return illegal_read_write;
      len = 48;
      break;
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      len = 64;
      break;
    default:  return illegal_format;
  }

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        if (strnicmp(xpqs[idx].prpName,Property,PROPERTY_NAME_SHORTSIZE)) continue;
        strncpy(data,xpqs[idx].prpDescription,len);
        return 0;
      }
    }
    if (el->EqmPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        if (strnicmp(prpqs[idx].prpName,Property,PROPERTY_NAME_SIZE)) continue;
        strncpy(data,prpqs[idx].prpDescription,MIN(len,PROPERTY_NAME_SIZE));
        return 0;
      }
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
    {
      if (strnicmp(prp->prpName,Property,PROPERTY_NAME_SIZE)) continue;
      strncpy(data,prp->prpDescription,len);
      return 0;
    }
    return illegal_property;
  }
  return non_existent_elem;
}
TINE_EXPORT int GetLocalHistoryData(char *eqm,char *prp, char *dev,DTYPE *dout,DTYPE *din)
{
  CONTRACT cn;
  strncpy(cn.EqmName,eqm,EQM_NAME_SHORTSIZE);
  strncpy(cn.EqmProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(cn.EqmDeviceName,dev,DEVICE_NAME_SIZE);
  cn.EqmSizeOut = dout->dArrayLength;
  cn.EqmFormatOut = BFMT(dout->dFormat);
  cn.EqmSizeIn = din->dArrayLength;
  cn.EqmFormatIn = BFMT(din->dFormat);
  return HISTQueryFunction(&cn,din->data.cptr,dout->data.cptr);
}
int HISTQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  SINT32 numread, nr;
  char *c, Property[PROPERTY_NAME_SIZE];
  time_t stop=time(NULL)-1; /* default values */
  double dstop = (double)stop, dstart = 0;
  float lwr=0,upr=0;
  int n,fmt,cc=0,index=0,sample=0,recsreadFlag=0,isSnapshot=0;
  SINT32 *lptr;
  double *dptr;
  ExportListStruct *el;
  PrpQueryStruct *prpqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  BYTE *dataptr = (BYTE *)data;

  /* check for server redirection from possible property query functions ... */
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL &&
        (n=el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs)) <= 0) ccerr(-n);
    if (el->EqmPropertyQueryFcn != NULL &&
        (n=el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs)) <= 0) ccerr(-n);
    break;
  }
  /* proceed ...*/
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".AR")) != NULL)  
  {
    *c = 0;
    return getRemoteHistoryServer(con->EqmName,Property);
  }
  if (strstr(Property,".HIST@") != NULL || strstr(Property,".HST@") != NULL)
    isSnapshot = TRUE;
  if ((c=strstr(Property,".HIST")) == NULL &&
      (c=strstr(Property,".HST")) == NULL) return illegal_property;
  *c = 0;
  if (con->EqmSizeOut == 0) return illegal_read_write;
  n = (int)con->EqmSizeOut;
  switch (fmt = LFMT(con->EqmFormatOut))
  {
    case CF_DOUBLE: case CF_FLOAT: case CF_LONG: case CF_INTFLTINT:
      recsreadFlag = 0;
      break;
    default:
      break;
  }
  if (con->EqmSizeIn > 0)
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_LONG:
        lptr = (SINT32 *)dataIn;
        switch (con->EqmSizeIn)
        {
          case 5: recsreadFlag = (int)lptr[4];
          case 4: sample = (int)lptr[3];
          case 3: index = (int)lptr[2];
          case 2: if (lptr[1] < (SINT32)stop) dstop = (double)lptr[1];
          case 1: dstart = (double)lptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_DOUBLE:
        dptr = (double *)dataIn;
        switch (con->EqmSizeIn)
        {
          case 5: recsreadFlag = (int)dptr[4];
          case 4: sample = (int)dptr[3];
          case 3: index = (int)dptr[2];
          case 2: if (dptr[1] < stop) dstop = dptr[1];
          case 1: dstart = dptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_FWINDOW:
        dstart = (double)((FWINDOW *)dataIn)[0].i1val;
        dstop = (double)((FWINDOW *)dataIn)[0].i2val;
        lwr = ((FWINDOW *)dataIn)[0].f1val;
        upr = ((FWINDOW *)dataIn)[0].f2val;
        if (lwr < upr) sample = -1;
        break;
      case CF_WINDOW:
        dstart = (double)((DWINDOW *)dataIn)[0].t1;
        dstop = (double)((DWINDOW *)dataIn)[0].t2;
        sample = ((DWINDOW *)dataIn)[0].i2val;
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    dstart = (double)getStartTimeFromDepth(con->EqmName,Property,n);
    if (dstart < 0) ccerr((int)(-dstart));
  }

  if (dstart < 0) dstart = 0;
  if (dstop < dstart) dstop = (double)time(NULL);
  if (isSnapshot) 
  { 
    if (dstart < 1.0) return argument_list_error;
    if (fmt > CF_FLOAT) return illegal_format;
    dstop = dstart;
    /* snapshot read doesn't use sample, index or recsread ... */
    sample = 0;
    index = 0;
    recsreadFlag = 0;
    lwr = upr = 0.0;
  }
  numread = getLTSData(con->EqmName,Property,con->EqmDeviceName,index,dstart,dstop,lwr,upr,
                       dataptr,fmt,n,sample,recsreadFlag);
  if (numread < 0) ccerr((int)(-numread));
  if (numread > 0) 
  {
    if (!isSnapshot && numread < (SINT32)con->EqmSizeOut - 1)
    {
      dataptr += numread * fmtsizeof(fmt);
      dstart = getDataTimeStamp() + 1.0;
      n -= numread;
    }
    else 
    {
      ccerr(0);
    }
  }

  nr = getSTSData(con->EqmName,Property,con->EqmDeviceName,index,dstart,dstop,lwr,upr,
                  dataptr,fmt,n,sample,recsreadFlag);
  if (nr < 0) ccerr((int)(-nr));
  numread += nr;
  if (numread < (SINT32)con->EqmSizeOut - 1) con->EqmSizeOut = numread;
err:
  return  cc;
}
#ifdef WIN32
int gCntNotPosted = 0;
int getNotPostedCount(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqmSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_SHORT:
      *(short *)data = (short)gCntNotPosted; break;
    case CF_LONG:
      *(int *)data = gCntNotPosted; break;
    case CF_FLOAT:
      *(float *)data = (float)gCntNotPosted; break;
    case CF_DOUBLE:
      *(double *)data = (double)gCntNotPosted; break;
    default: return illegal_format;
  }
  return 0;

}
#endif
StockPropertyStruct StockProperty[] =
{
     {"PROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE|AT_CHANNEL,"","Exported Properties",propsQuery,FALSE},
     {"PROPS",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE|AT_CHANNEL,"","Exported Properties",propsQuery,FALSE},
     {"PROPS",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery,FALSE},
     {"PROPS",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE},
     {"PROPS",20,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE},
     {"PROPS",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE},
     {"PROPS",10,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE},
     {"PROPS",12,1,CF_STRUCT,CF_NAME64,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE},
     {"PROPS",12,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE},
     {"NPROPS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery,FALSE},
     {"PROPERTIES",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",20,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",10,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",12,1,CF_STRUCT,CF_NAME64,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE},
     {"PROPERTIES",12,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE},
     {"NPROPERTIES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery,FALSE},
     {"DEVICES",512,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Device Names",deviceQuery,FALSE},
     {"DEVICES",512,1,CF_NAME16,CF_NAME16,CA_READ,AT_SINGLE,"","Device Names",deviceQuery,FALSE},
     {"NDEVICES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Devices",ndeviceQuery,FALSE},
     {"ALARMS",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Total Current Alarm List",alarmsQuery,TRUE},
     {"ALARMS",256,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMSr4","Total Current Alarm List",alarmsQuery,TRUE},
     {"NALARMS",5,0,CF_LONG,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Current Alarm Information",nalarmsQuery,FALSE},
     {"ALMDEFS",48,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"ADSr4","Defined Alarm Descriptions",alarmDefsQuery,FALSE},
     {"ALMDEFS",32,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"ADS","Defined Alarm Descriptions",alarmDefsQuery,FALSE},
     {"NALMDEFS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Defined Alarms",nalarmDefsQuery,FALSE},
     {"ALARMSEXT",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Specified Current Alarm List",alarmsQuery,TRUE}, 
     {"ALARMSEXT",256,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMSr4","Specified Current Alarm List",alarmsQuery,TRUE}, 
     {"NALMWATCH",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of alarms in watch table",nalarmWatchQuery,FALSE},
     {"ALMWATCHTBL",32,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AWSr4","Current Alarm Watch Table",alarmWatchQuery,FALSE}, 
     {"MESSAGE",0,256,CF_NULL,CF_TEXT,CA_WRITE,AT_SINGLE,"","Accept Message",msgDisplay,FALSE},
     {"ACTIVITY",68,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"","Current Server Activity",activityQuery,FALSE},  
     {"ACTIVITY",1,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AQS","Current Server Activity",activityQuery,FALSE},  
     {"CONTRACTS",920,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Contract List",contractsQuery,FALSE},
     {"CONTRACTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CONQS","Current Contract List",contractsQuery,FALSE},
     {"CONTRACTS",12,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CTQSr4","Current Contract List",contractsQuery,FALSE},
     {"CLIENTS",540,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Client List",clientsQuery,FALSE},
     {"CLIENTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CLNQS","Current Client List",clientsQuery,FALSE},
     {"SRVEXIT",0,0,CF_NULL,CF_NULL,CA_WRITE,AT_UNKNOWN,"","Stop Server (when enabled)",srvExit,FALSE},
     {"ACCESSLOCK",0,2,CF_NULL,CF_SHORT,CA_WRITE,AT_UNKNOWN,"","lock type/lock duration",lockAccess,FALSE},
     {"ACCESSLOCK",2,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","lock owner/address",lockAccess,FALSE},
     {"ADDUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add User to Access List",addRegisteredUser,FALSE},
     {"DELUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Remove User from Access List",rmvRegisteredUser,FALSE},
     {"USERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read User Access List",getRegisteredUserList,FALSE},
     {"DENIEDUSERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read Denied User List",getDeniedUserList,FALSE},
     {"NUSERS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Users with Access",getNumRegisteredUsers,FALSE},
     {"STOCKPROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE},
     {"STOCKPROPS",68*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE},
     {"STOCKPROPS",188*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE},
     {"STOCKPROPS",188*10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE},
     {"STOCKPROPS",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Stock Properties",stockpropsQuery,FALSE},
     {"STOCKPROPS",12,64,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PRPQSr4","Stock Properties",stockpropsQuery,FALSE},
     {"NSTOCKPROPS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Stock Properties",nstockpropsQuery,FALSE},
     {"IPNETS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read IP Access List",getRegisteredIPNetList,FALSE},
     {"NIPNETS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of IP Nets with Access",getNumRegisteredIPNets,FALSE},
     {"ADDIPNET",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add IP Net to Access List",addRegisteredIPNet,FALSE},
     {"IPXNETS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read IPX Access List",getRegisteredIPXNetList,FALSE},
     {"NIPXNETS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of IPX Nets with Access",getNumRegisteredIPXNets,FALSE},
     {"DEBUGLEVEL",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","get debug level",setDebugLevel,FALSE},
     {"DEBUGLEVEL",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","get/set debug level",setDebugLevel,FALSE},
     {"LOGCOMMANDS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","read en/dis-able command logging flag",setLogCommands,FALSE},
     {"LOGCOMMANDS",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","en/dis-able command logging",setLogCommands,FALSE},
     {"LOGDEPTH",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","get log depth in lines",setLogDepth,FALSE},
     {"LOGDEPTH",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","get/set log depth in lines",setLogDepth,TRUE},
     {"LOGFILE",8000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get log file",getFecLog,FALSE},
     {"SRVVERSION",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","TINE Server Version Number",getTINEVersion,FALSE},
     {"SRVOS",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","General Server OS",getServerOS,FALSE},
     {"SRVDESC",64,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Description",getServerDescription,FALSE},
     {"SRVLOCATION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Location",getServerLocation,FALSE},
     {"SRVSTATS",16,0,CF_LONG,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Server Statistics counters",getServerStats,FALSE},
     {"SRVALIASLIST",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Alias List (Name/Alias pairs)",getServerAliasList,FALSE},
     {"NHISTORIES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of history properties",nhistoryRecordsQuery,FALSE},
     {"HISTORIES",32,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","history properties",historyRecordsQuery,FALSE},
     {"ADDHISTORY",0,1,CF_NULL,CF_STRUCT,CA_WRITE,AT_SINGLE,"HQS","add/edit a history record",historyRecordsAppend,FALSE},
     {"STRUCTFORMAT",512,1,CF_NAME64DBLDBL,CF_NAME16,CA_READ,AT_SINGLE,"","Registered structure information",taggedStructQuery,FALSE},
     {"BITFIELDFORMAT",512,1,CF_NAME16I,CF_NAME16,CA_READ,AT_SINGLE,"","Registered bitfield information",bitfieldStructQuery,FALSE},
     {"SRVSTARTTIME",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime,FALSE},
     {"SRVSTARTTIME",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime,FALSE},
     {"APPDATE",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile Time",getAppDate,FALSE},
     {"APPDATE",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile  Time",getAppDate,FALSE},
     {"APPVERSION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Version Number",getAppVersion,FALSE},
     {"SRVSELFTEST",32000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get self-test file",getSelfTestFile,FALSE},
     {"SRVLOGFILES",32,0,CF_NAME64,CF_NULL,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList,FALSE},
     {"SRVLOGFILES",32,64,CF_NAME64,CF_TEXT,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList,FALSE},
     {"SRVLOGFILE",32000,0,CF_TEXT,CF_TEXT,CA_READ|CA_WRITE,AT_SCALAR,"","get specific log file ",getFecLogFile,TRUE},
     {"SRVBINFILE",32000,0,CF_BYTE,CF_TEXT,CA_READ|CA_WRITE,AT_SCALAR,"","get specific binary file ",getFecFile,TRUE},
     {"SRVLASTACCESS",5,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","get most recent access",getLastAccessList,FALSE},
     {"SRVADDR",5,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","get server address parameters",getSrvAddressParams,FALSE},
     {"SRVRESET",0,1,CF_NULL,CF_LONG,CA_WRITE,AT_UNKNOWN,"","Reset Server (when enabled)",srvReset,FALSE},
     {"SRVINIT",0,0,CF_NULL,CF_NULL,CA_WRITE,AT_UNKNOWN,"","Re-initialize Server (when enabled)",srvInit,FALSE},
#    ifdef WIN32
     {"SRVNOTPOSTED",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Number Contracts not posted",getNotPostedCount,FALSE},
#    endif
};
int NrStockProperties = (sizeof(StockProperty)/sizeof(StockPropertyStruct));

HashXRefStruct **StockPropXRefTable = NULL;

int makeStockPropXRefTable(void)
{
  int k,idx;
  HashXRefStruct *xref;

  if ((StockPropXRefTable=(HashXRefStruct **)SystemCalloc(STOCKPROP_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_server_memory;
  for (k=0; k<NrStockProperties; k++)
  {
    idx = (int)(ElfHash((unsigned char *)StockProperty[k].prpName) % STOCKPROP_HASH_SIZE);
    for (xref=StockPropXRefTable[idx]; xref && xref->index != k; xref = xref->next)
    {
      if (!strnicmp(StockProperty[k].prpName,StockProperty[xref->index].prpName,PROPERTY_NAME_SHORTSIZE)) break;
    }
    if (xref == NULL) 
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
      xref->index = k;
      xref->next = StockPropXRefTable[idx];
      StockPropXRefTable[idx] = xref;
    }
  }
  return 0;
}
StockModifierStruct StockModifier[] =
{
  {".EGU",EGUQueryFunction,FALSE},
  {".HIST",HISTQueryFunction,TRUE},
  {".HST",HISTQueryFunction,TRUE},
  {".HIST@",HISTQueryFunction,TRUE},
  {".HST@",HISTQueryFunction,TRUE},
  {".ARCH",HISTQueryFunction,TRUE},
  {".ARC",HISTQueryFunction,TRUE},
  {".AR",HISTQueryFunction,TRUE},
  {".NAM",NAMQueryFunction,FALSE},
  {".DESC",DESCQueryFunction,FALSE},
  {".DSC",DESCQueryFunction,FALSE},
  {".BIT.",gateFunction,FALSE},
  {".MASK.",gateFunction,FALSE},
  {".GATE.",gateFunction,FALSE},
};
int NrStockModifiers = (sizeof(StockModifier)/sizeof(StockModifierStruct));

short IsStockProperty(char *Property)
{
  if (strchr(Property,'*') != NULL) return STOCK_PROP; /* maps to "PROPERTIES" */
  return GetStockPropertyId(Property) < 0 ? EMPTY_PROP : STOCK_PROP;
}
short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty)
{
  int i;
  char *c;
  static char metaPropertyString[64];

  strncpy(metaPropertyString,Property,PROPERTY_NAME_SIZE);
  for (i=0; i<NrStockModifiers; i++)
  {
    if (!strncmp(StockModifier[i].modName,".AR",2)) continue; /* these don't count ! */
    if ((c=strstr(metaPropertyString,StockModifier[i].modName)) != NULL)
    {
      *c++ = 0;
      if (baseProperty != NULL) *baseProperty = metaPropertyString;
      if (metaProperty != NULL) *metaProperty = c;
      return TRUE;
    }
  }
  return FALSE;
}

short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate)
{
  char *c;
  int bit;

  if (baseProperty == NULL || gate == NULL) return FALSE;
  if (LFMT(con->EqmFormatOut) != CF_BYTE &&
      LFMT(con->EqmFormatOut) != CF_SHORT &&
      LFMT(con->EqmFormatOut) != CF_LONG) return FALSE;
  strncpy(baseProperty,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(baseProperty,".BIT.")) != NULL)
  {
    *c = 0;
    c += 5;
    bit = atoi(c);
    if (bit < 0 || bit >= 8*fmtsizeof(con->EqmFormatOut)) return FALSE;
    *gate = 0x01 << bit;
    return GATE_BOOLEAN;
  }
  if ((c=strstr(baseProperty,".MASK.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return GATE_BOOLEAN;
  }
  if ((c=strstr(baseProperty,".GATE.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return GATE_FILTER;
  }
  return FALSE;
}
