/** \file ttycmd.c
 * \brief TINE console command line API
 *
 * The API routines offered here allow the developer to register his own command-line 
 * calling routines to be used for interacting with a running server at the command line.
 *
 */

#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stddef.h>
#include <malloc.h>
#include "srvcore.h"
#include "dbtools.h"
#include "srvip.h"
#include "globals.h"
#include "ttycmd.h"

extern int ALMdebug;

/* command relevant entries: */
void ExitEQPModules(void);

#define CMDBUFSIZE 20

#if defined(WIN32)
#include <windows.h>
#ifndef UNDER_CE
#include <conio.h>
#endif
size_t IPCSocket = 0;
void stopPipeListener(void)
{
# ifndef NO_SERVICE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  if (pipeInst == NULL) return;
  feclog("shutting down service pipe");
  DisconnectNamedPipe(pipeInst->hPipeInst);
  IPCSocket = 0; 
# endif
}
void resetPipeListener(PIPEINST *pipeInst)
{
# ifndef NO_SERVICE_PIPE
  if (pipeInst == NULL) return;
  DisconnectNamedPipe(pipeInst->hPipeInst);
  pipeInst->oOverlap.hEvent = pipeInst->hPipeConnectEvent;
  ConnectNamedPipe(pipeInst->hPipeInst,&pipeInst->oOverlap);
  pipeInst->isConnected = FALSE;
# else
  pipeInst = pipeInst;
# endif
}
int ipcprintf(char *str)
{
# ifndef NO_SERVICE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  int i,rc,slen=strlen(str),wlen=0;
  char *sptr = str;
  if (IPCSocket != 0) 
  {
    for (i=0; i<3; i++)
    {
      rc = WriteFile(pipeInst->hPipeInst,sptr,(UINT32)slen,&wlen,NULL);
      if (rc == 0) 
      {
        rc = GetLastError();
        if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
        break;
      }
      if (wlen < slen) 
      {
        sptr = &str[wlen];
        slen -= wlen;
      }
      else
      {  
        break;
      }
    }
  }
# endif
  return 0;
}
void GetCommand(void)
{
  static char prvcmd[CMDBUFSIZE][CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen[CMDBUFSIZE], clen = 0, prvptr = 0, curptr = 0;
  int i;
  char c;
  if (kbhit())
  {
    switch (c=getch())
    {
      case -32: /* extended (cursor movement) */
        switch (c=getch())
        {
          case 'M': /* right */
                    if (clen > len)
                    {
                      putch(cmdbuf[len++]);
                    }
                    break;
          case 'S': /* Delete */
          case 'K': /* left */
                    if (len > 0)
                    {
                      putch('\r'); putch('>'); 
                      for (i=0; i<len; i++) putch(' ');
                      cmdbuf[len--] = 0;
                      putch('\r'); putch('>'); 
                      for (i=0; i<len; i++) putch(cmdbuf[i]);
                    }
                    break;
          case 'H': /* up */
                    putch('\r'); putch('>'); 
                    for (i=0; i<len; i++) putch(' ');
                    strncpy(cmdbuf,prvcmd[prvptr],len=prvlen[prvptr]);
                    cmdbuf[len] = 0;
                    putch('\r'); putch('>'); 
                    printf(cmdbuf);
                    prvptr = (CMDBUFSIZE + prvptr - 1) % CMDBUFSIZE;
                    break;
          case 'P': /* down */
                    if (clen > len)
                    {
                      cmdbuf[clen] = 0;
                      putch('\r'); putch('>'); 
                      printf(&cmdbuf[len]);
                      len = clen;
                    }
                    break;
        }

        break;
      case 9: /* tab */
        break;
      case 8: /* backspace */
        if (len > 0)
        {
          putch('\r'); putch('>'); 
          for (i=0; i<len; i++) putch(' ');
          cmdbuf[len--] = 0;
          putch('\r'); putch('>'); 
          for (i=0; i<len; i++) putch(cmdbuf[i]);
        }
        break;
      case 10:
      case 13:
        ttyoutput("");
        if (len == 0) break;
        cmdbuf[len] = 0;
        InterpretCommand(cmdbuf);
        CMDcount++;
        curptr = (curptr + 1) % CMDBUFSIZE;
        prvptr = curptr;
        strncpy(prvcmd[prvptr],cmdbuf,len);
        prvlen[prvptr] = len;
        len = 0;
        break;
      default:
        putch(c);
        cmdbuf[len++] = c;
        if (len > CMDSIZE-2) len = 0;
        clen = MAX(clen,len);
        break;
    }
  }
  return;
}
int GetCommandEx(size_t fd,char *buf,int bufsiz)
{
# ifndef NO_SERVICE_PIPE
  int rc, siz, len=0;
  PIPEINST *pipeInst = (PIPEINST *)fd;

  if ((rc=PeekNamedPipe(pipeInst->hPipeInst,NULL,0,NULL,&siz,NULL)) == 0)
  {
    resetPipeListener(pipeInst);
    return -1;
  }
  siz = MIN(siz,bufsiz);
  if ((rc=ReadFile(pipeInst->hPipeInst,pipeInst->chBuf,siz,&len,&pipeInst->oOverlap)) == 0) 
  {
    rc = GetLastError();
    if (rc == ERROR_IO_PENDING) return 0;
    resetPipeListener(pipeInst);
    return -1;
  }
  strncpy(buf,pipeInst->chBuf,len);
  memset(pipeInst->chBuf,0,bufsiz);
  buf[len] = 0;
  if (len == 0) 
  {
    return 0;
  }
  CMDcount++;
  return len;
# else
  fd = fd; buf = buf; bufsiz = bufsiz;
  return 0;
# endif
}
extern int isWindowsNT;
size_t CMDlisten(const char *name)
{
# ifndef NO_SERVICE_PIPE
  PIPEINST *pipeInst=NULL; 
  HANDLE hPipeConnectEvent,hPipeReadEvent;
  int rc, cc = 0;
  int openMode = PIPE_ACCESS_DUPLEX|FILE_FLAG_OVERLAPPED;
  int pipeMode = PIPE_TYPE_MESSAGE|PIPE_READMODE_MESSAGE|PIPE_WAIT;
  char *ptr = getenv("TINE_REMOTE_PIPE");

  if (!isWindowsNT)
  {
    feclog("Remote PIPE disabled (not a Windows NT OS)");
    return 0;
  }
  if (ptr != NULL && !stricmp(ptr,"NO"))
  {
    feclog("Remote PIPE disabled via environment");
    return 0;
  }

  /* no security attribute, manual-reset event, initial state = signaled, unnamed event object */
  hPipeConnectEvent = CreateEvent(NULL,TRUE,TRUE,NULL);
  if (hPipeConnectEvent == NULL) ccerr(code_failure); 
  hPipeReadEvent = CreateEvent(NULL,TRUE,TRUE,NULL);
  if (hPipeReadEvent == NULL) ccerr(code_failure); 
  if ((pipeInst=(PIPEINST *)calloc(1,sizeof(PIPEINST))) == NULL) ccerr(out_of_server_memory);
  pipeInst->hPipeConnectEvent = hPipeConnectEvent; 
  pipeInst->hPipeReadEvent = hPipeReadEvent;     
  pipeInst->oOverlap.hEvent = pipeInst->hPipeConnectEvent; 
  pipeInst->hPipeInst = 
    CreateNamedPipe(name,openMode,pipeMode,MAXIPCCLIENTS,IPCBUFSIZE,IPCBUFSIZE,1000,NULL);
  if (pipeInst->hPipeInst == INVALID_HANDLE_VALUE) ccerr(code_failure);
  if (ConnectNamedPipe(pipeInst->hPipeInst,&pipeInst->oOverlap) != 0) ccerr(code_failure);
  pipeInst->isConnected = FALSE; 
err:
  if (cc != 0) 
  {
    rc = GetLastError();
    feclog("CMD listen : %s (windows err %d)",erlst[cc],rc);
  }
  return (size_t)pipeInst;
# else
  name = name;
  return 0;
# endif
}
size_t CMDaccept(PIPEINST *pipeInst)
{
  if (pipeInst == NULL) return 0;
  /* just test the wait status */
  if (WaitForSingleObject(pipeInst->oOverlap.hEvent,0) != 0) return 0;
  if (pipeInst->isConnected == FALSE)
  {
    pipeInst->isConnected = TRUE;
    pipeInst->oOverlap.hEvent = pipeInst->hPipeReadEvent;     
  }
  return (size_t)pipeInst;
}
#elif defined(UNIX)
#include <sys/un.h>
int  ipcfd = -1;
int  IPCSocket = 0;
int  IPCfds[MAXIPCCLIENTS],nipcclients = 0;
int ipcprintf(char *str)
{
# ifndef NO_SERVICE_PIPE
  int i;
  for (i=0; i<nipcclients; i++) write(IPCfds[i],str,strlen(str));
# endif
  return 0;
}
int GetCommandEx(int fd,char *buf,int bufsiz)
{
  static int len=0;
  static char prvcmd[CMDSIZE];

  if ((len=read(fd,buf,bufsiz)) < 0) return -1;
  buf[len] = 0;
  if (len == 0) 
  {
    /* don't let it spin in the background ! */
    if (fd == STDIN_FILENO) foregroundTTY = 0;
    return 0;
  }
  if (strcmp(buf,"!")) 
  { /* make a note of this command */
    strncpy(prvcmd,buf,CMDSIZE);
  }
  else
  { /* re-issue the previous command */
    strncpy(buf,prvcmd,CMDSIZE);
  }
  if (fd != STDIN_FILENO) { ipcfd = fd; printf(buf); }
  else ipcfd = -1;
  CMDcount++;
  return len;
}
int CMDlisten(const char *name)
{
  int fd,len;
  struct sockaddr_un unix_addr;
# ifdef SVR4
  int fdarr[2];
# endif

  unlink(name);

# ifndef SVR4
  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path,name);
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) return -1;
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) return -2;
  if (listen(fd, 5) < 0) return -3;
# else
  if ((fd=creat(name,FIFO_MODE)) < 0) return -1;
  if (close(fd) < 0) return -2;
  if (pipe(fdarr) < 0) return -3;
  if (ioctl(fdarr[1],I_PUSH,"connld") < 0) return -4;
  if (fattach(fdarr[1],name) < 0) return -5;
  fd = fdarr[0];
# endif /* SVR4 */

  return fd;
}
int CMDconnect(const char *name)
{
  int cc = 0, fd = 0,len;
  struct sockaddr_un unix_addr;
# ifndef SVR4
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) ccerr(-1);
  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  sprintf(unix_addr.sun_path,
    strlen(CLN_PATH) == 9 ? "%s%05d" : "%s%09d", CLN_PATH, getpid());

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if (len != 16) ccerr(-5);
# endif
  unlink(unix_addr.sun_path);
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) ccerr(-2);
  if (chmod(unix_addr.sun_path, CLN_PERM) < 0) ccerr(-3);
  memset(&unix_addr, 0, sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, name);
# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
  strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
# endif
  if (connect(fd, (struct sockaddr *)&unix_addr,len) < 0) ccerr(-4);
#else
  if ((fd=open(name,O_RDWR)) < 0) ccerr(-1);
  if (isastream(fd) == 0) 
  {
    close(fd);
    ccerr(-2);
  }
#endif /* SVR4 */
err:
  if (cc)
  {
    if (fd > 0) close(fd);
  }
  return fd;
}

#define STALE 30
int CMDaccept(int lsnfd, uid_t *uidptr)
{
  int cc=0,clnfd,len;
  time_t staletime;
  char pipename[15];
# ifndef SVR4
  struct sockaddr_un unix_addr;
  struct stat statbuf;
# else
  struct strrecvfd recvfd;
# endif

# ifndef SVR4
  len = sizeof(unix_addr);
  if ((clnfd=accept(lsnfd, (struct sockaddr *)&unix_addr,&len)) < 0) ccerr(-1);
  len -= sizeof(unix_addr.sun_family);

  strncpy(pipename,unix_addr.sun_path,14); pipename[14] = 0;
  if (stat(pipename,&statbuf) < 0) ccerr(-2);

# ifdef S_ISSOCK
  if (S_ISSOCK(statbuf.st_mode) == 0) ccerr(-3);
# endif

  if ((statbuf.st_mode & (S_IRWXG | S_IRWXO)) ||
      (statbuf.st_mode &  S_IRWXU) != S_IRWXU) ccerr(-4);

  staletime = time(NULL) - STALE;
  if (statbuf.st_atime < staletime ||
      statbuf.st_ctime < staletime ||
      statbuf.st_mtime < staletime) ccerr(-5);

  if (uidptr != NULL) *uidptr = statbuf.st_uid;

  unlink(pipename);
# else
  if (ioctl(lsnfd,I_RECVFD,&recvfd) < 0) ccerr(-1);
  if (uidptr != NULL) *uidptr = recvfd.uid;
  clnfd = recvfd.fd;
# endif /* SVR4 */
  return clnfd;
err:
  perror("CMDaccept");
  if (cc < -1) printf("name %s len %d\n",pipename,len);
  return cc;
}

#define IPCquit(txt) { perror(txt); printf("return %d\n",clnfd); return; }
void addIPCcln(void)
{
  int clnfd;
  uid_t uid;
  char strbuf[CMDSIZE];

  if ((clnfd = CMDaccept(IPCSocket,&uid)) < 0) IPCquit("accept error");
  if (nipcclients < MAXIPCCLIENTS)
  {
    IPCfds[nipcclients] = clnfd;
    nipcclients++;
    printf("Remote Session granted to uid %d\n",uid);
  }
  else
  {
    sprintf(strbuf,"Maximum %d clients exceeded\n",MAXIPCCLIENTS);
    write(clnfd,strbuf,strlen(strbuf));
  }
}
#ifndef BSDSIGNAL
Sigfunc *bsdsignal(int signo, Sigfunc *disp)
{
  struct sigaction act;

  act.sa_handler = (void(*)())disp;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;

  if (signo != SIGALRM) act.sa_flags |= SA_RESTART;
  return sigaction(signo,&act,NULL) < 0 ? SIG_ERR : 0;
}
#endif

void sig_pipe(int signo)
{
  printf("Remote session terminated\n");
#ifdef INTEL
  if (bsdsignal(SIGPIPE, sig_pipe) == SIG_ERR) printf("signal error\n");
#endif
}

#elif defined(DOS)
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
void GetCommand(void)
{
  static char prvcmd[CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen = 0, clen = 0;
  int i,x,y;
  char c;

  if (kbhit())
  {
    switch (c=getch())
    {
      case 0: /* extended (cursor movement) */
        switch (c=getch())
        {
          case 'M': /* right */
                    if (clen > len)
                    {
                      putch(cmdbuf[len++]);
                    }
                    break;
          case 'K': /* left */
                    if (len > 0)
                    {
                      x = wherex(); y = wherey();
                      gotoxy(x-1,y); putch(' ');
                      gotoxy(x-1,y);
                      len--;
                    }
                    break;
          case 'H': /* up */
                    x = wherex(); y = wherey();
                    for (i=len-1; i>=0; i--) { gotoxy(i,y); putch(' '); }
                    gotoxy(0,y);
                    strncpy(cmdbuf,prvcmd,len=prvlen);
                    cmdbuf[len] = 0;
                    printf(cmdbuf);
                    break;
          case 'P': /* down */
                    if (clen > len)
                    {
                      cmdbuf[clen] = 0;
                      printf(&cmdbuf[len]);
                      len = clen;
                    }
                    break;
        }
        break;
      case 9: /* tab */
        break;
      case 8: /* backspace */
        if (len > 0)
        {
          x = wherex(); y = wherey();
          gotoxy(x-1,y); putch(' ');
          gotoxy(x-1,y);
          len--;
        }
        break;
      case 10:
      case 13:
        ttyoutput("");
        cmdbuf[len] = 0;
        InterpretCommand(cmdbuf);
        CMDcount++;
        strncpy(prvcmd,cmdbuf,len);
        prvlen = len;
        len = 0;
        break;
      default:
        putch(c);
        cmdbuf[len++] = c;
        if (len > CMDSIZE-2) len = 0;
        clen = MAX(clen,len);
        break;
    }
  }
  return;
}

#elif defined(VXWORKS)
int ipcprintf(char *str)
{
  str = str;
  return 0;
}

#elif defined(NIOS)
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
int putch(int c)
{
# ifdef NIOS_I
  nr_txchar(c);
# endif
  return c;
}
int getch(void)
{
# ifdef NIOS_I
  return nr_rxchar();
# else
  return 0;
# endif
}
void GetCommand(void)
{
  static char prvcmd[CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen = 0, clen = 0;
  int c;

  switch (c=getch())
  {
    case 27:
    case 0:
      if ((c=getch()) != '[') break;
      switch (c=getch())
      {
        case 'C': /* right */
                  if (clen > len)
                  {
                    putch(cmdbuf[len++]);
                  }
                  break;
        case 'D': /* left */
                  if (len > 0)
                  {
                    len--;
                  }
                  break;
        case 'A': /* up */
                  strncpy(cmdbuf,prvcmd,len=prvlen);
                  cmdbuf[len] = 0;
                  printf(cmdbuf);
                  break;
        case 'B': /* down */
                  if (clen > len)
                  {
                    cmdbuf[clen] = 0;
                    printf(&cmdbuf[len]);
                    len = clen;
                  }
                  break;
      }
      break;
    case 9: /* tab */
      break;
    case 8: /* backspace */
      if (len > 0)
      {
        len--;
      }
      break;
    case 10:
    case 13:
      ttyoutput("");
      cmdbuf[len] = 0;
      InterpretCommand(cmdbuf);
      CMDcount++;
      strncpy(prvcmd,cmdbuf,len);
      prvlen = len;
      len = 0;
      break;
    default:
      if (c < 0) return;
      putch(c);
      cmdbuf[len++] = c;
      if (len > CMDSIZE-2) len = 0;
      clen = MAX(clen,len);
      break;
  }
  return;
}

#else
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
void execCmd(char *cmdbuf)
{
  InterpretCommand(cmdbuf);
}
#endif

struct CommandListStruct {
  char cmd[10];
  int (*fcn)(char *,char *);
};
struct CommandListStruct CommandList[] = {
  { "quit",quit},
  { "exit",quit},
  { "help",help},
  { "set",setVal},
  { "get",getVal}
};
int NumCommands = sizeof(CommandList)/sizeof(struct CommandListStruct);
int CommandLevel = 0;

USRCMD *usrcmd = NULL;
int numUsrCmd = 0;

USRFCN *usrfcn = NULL;
int numUsrFcn = 0;

int InterpretCommand(char *buf)
{
  char cmd[CMDSIZE],strbuf[CMDSIZE];
  char *c,*ce=NULL,*cp=NULL,*arg1=NULL, *arg2=NULL;
  int cc=0,i,len;
  int argc;
  char *argv[10],vstr[32],sstr[128];
  USRFCN *uf;

  if (buf == NULL || strlen(buf) <= 1) ccerr(0);
  strncpy(cmd,buf,CMDSIZE-1); cmd[CMDSIZE-1] = 0;
  for (i=0; i<NumCommands; i++)
   if (!strncmp(cmd,CommandList[i].cmd,strlen(CommandList[i].cmd))) break;

  if (i < NumCommands) 
  {
    if ((c=strchr(cmd,'=')) != NULL)
    {
      arg2 = c+1;
      *c = 0;
    }
    arg1 = &cmd[strlen(CommandList[i].cmd)+1];
    ccerr(CommandList[i].fcn(arg1,arg2));
  }
  for (i=0; i<32 && cmd[i] == ' '; i++);
  strncpy(vstr,&cmd[i],32);
  if ((c=strchr(vstr,'(')) != NULL) *c++ = 0;
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    if (uf->name != NULL && !strncmp(vstr,uf->name,32))
    {
      argc = 0;
      len = (int)strlen(uf->name);
      if (c != NULL && (ce=strchr(c,')')) != NULL)
      {
        *ce = 0; strncpy(sstr,c,62); sstr[63] = 0;
        for (i=0,c=strtok_r(sstr,",",&cp); c && i<10; i++,c=strtok_r(NULL,",",&cp))
        {
          argv[i] = c;
        }
        argc = i;
      }
      i = uf->fcn(argc,argv);
      sprintf(strbuf,"result is %d",i);
      ttyoutput(strbuf);
      ccerr(0);  
    }
  }
err: 
  if (cc) ttyoutput("\n>Syntax: set name = value or get name");
  ttyoutput("");
  return cc;
}
int help(char *s1,char *s2)
{
  USRCMD *uc = NULL;
  USRFCN *uf = NULL;
  char sstr[64],strbuf[CMDSIZE];

  ttyoutput("Currently available commands:");
  ttyoutput("\tquit                  - terminates process (foreground) or service viewer (background)");
  ttyoutput("\tkill                  - terminates process as well as service viewer");
  if (gEqmListSize)
  {
    ttyoutput("\tget modules           - displays a list of registered equipment modules");
    ttyoutput("\tget properties(<eqm>) - displays the registered properties for equipment module <eqm>");
    ttyoutput("\tget devices(<eqm>)    - displays the registered devices for equipment module <eqm>");
    ttyoutput("\tget clients           - displays the current consumer list");
    ttyoutput("\tget contracts         - displays the current contract list");
    ttyoutput("\tget contract(#)       - displays contract Nr. <#> information");
  }
  if (nglobals)
  {
    ttyoutput("\tget globals           - displays the current globals list");
  }
  if (nConnectionTableEntries)
  {
    ttyoutput("\tget connections        - displays the current connection list");
    ttyoutput("\tget connection(#)      - displays connection Nr. <#> information");
  }
  ttyoutput("\tget BurstLimit         - displays Burst Limit in packets");
  ttyoutput("\tset BurstLimit         - sets Burst Limit to num packets input");
  ttyoutput("\tget CycleDelay         - displays Cycle Delay in msec");
  ttyoutput("\tset CycleDelay         - sets Cycle Delay to msec input");
# ifdef DOS
  ttyoutput("\tget memory             - displays available memory");
  ttyoutput("\tget watchdog           - displays current watchdog settings");
  ttyoutput("\tset watchdog           - enables (non zero value)/disables (zero) watchdog");
# endif
  ttyoutput("\tget time               - displays local time");
  ttyoutput("\tget version            - displays TINE version number");
  ttyoutput("\tget users              - displays WRITE access user list");
  ttyoutput("\tget nets               - displays WRITE access net list");
  ttyoutput("\tget stats              - displays operational statistics");
  ttyoutput("\tget users              - displays the users with WRITE permission");
  ttyoutput("\tget nets               - displays the networks with WRITE access");
  ttyoutput("\tget filter             - displays current debug output filter string");
  ttyoutput("\tset filter             - sets debug output filter string");
  ttyoutput("\tset debug = 0          - turns debug printing off");
  ttyoutput("\tset debug = 1          - sets debug level 1 (trace RPC commands)");
  ttyoutput("\tset debug = 2          - sets debug level 2 (trace network activity)");
  ttyoutput("\tset debug = 3          - sets debug level 2 (trace data exchange)");
  ttyoutput("\tset debug = 4          - sets debug level 3 (full diagnostic trace )");
  ttyoutput("\tset logdbg = 0         - turns debug logging off");
  ttyoutput("\tset logdbg = 1         - turns debug logging on");
  ttyoutput("\thelp                   - display this list");
  ttyoutput("\n>Extra commands:");
  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    sstr[0] = 0;
    if (uc->access & CA_READ) strcpy(sstr,"get ");
    if (uc->access & CA_WRITE)
    {
      if (strlen(sstr)) strcat(sstr," or set ");
      else strcpy(sstr,"set ");
    }
    if (uc->iparam != NULL) strcat(sstr,"integer value");
    else if (uc->fparam != NULL) strcat(sstr,"float value");
    else if (uc->fcn != NULL) strcat(sstr,"function value");
    sprintf(strbuf,"\t%s\t - %s",uc->cmd,sstr);
    ttyoutput(strbuf);
  }
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    sstr[0] = 0;
    sprintf(strbuf,"\t%s()",uf->name);
    ttyoutput(strbuf);
  }
  return 0;
}
int quit(char *s1,char *s2)
{
  ExitEQPModules();
  quitServer();                   /* non-servers must cover this routine */
  if (nConnectionTableEntries) closeClients(TRUE);      /* non-clients must cover this routine */
# ifdef DOS
  disableWatchDog();
# endif
  feclog("stopped");
  exit(0);
  return 0;
}
int getVal(char *var,char *s)
{
  USRCMD *uc;
  char *c,*arg=NULL,*cp=NULL,sstr[128],vstr[32],strbuf[CMDSIZE];
  int i,index=-1,len,ival[4];

  if (var == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL || (c=strchr(vstr,'[')) != NULL) 
  {
    *c++ = 0;
    index = atoi(c);
    arg = c;
    if ((c=strchr(arg,')')) != NULL || (c=strchr(arg,']')) != NULL) 
    { /* well formed */
      *c = 0;
      if ((c=strchr(arg,'"')) != NULL)
      { /* explicitly entered as a string */
        arg++;
        if ((c=strchr(arg,'"')) != NULL)
        { /* well formed string */
          *c = 0;
        }
      }
    }
    strtrm(arg);
  }

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_READ)
    {
      if (uc->iparam != NULL)
      {
        sprintf(strbuf,"%s is %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = (int)strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;  
      }
    }
  }
  if (strstr(var,"filter"))
  {
    sprintf(strbuf,"debug name filter : %s",tagNameFilter);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"version")) return dumpVersion();
  if (strstr(var,"semaphores")) return dumpSemaphores();
  if (strstr(var,"debug"))
  {
    sprintf(strbuf,"Debug level %d",tineDebug);
    ttyoutput(strbuf);
    return 0;
  }
# ifdef DOS
  if (strstr(var,"watchdog"))
  {
    sprintf(strbuf,"hardware watchdog %s",
            !watchdogBaseAddress ? "not configured" :
            watchdogEnabled ? "enabled" : "disabled");
    ttyoutput(strbuf);
    return 0;
  }
# endif
  if (strstr(var,"time")) return dumpServerTime();
  if (strstr(var,"users")) { dumpRegisteredUsers(); return 0; }
  if (strstr(var,"nets")) { dumpIPNets(); return 0; }
  if (strstr(var,"globals")) { dumpGlobals(); return 0; }
  if (strstr(var,"CommandLevel"))
  {
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  if (gEqmListSize)
  {
    if (strstr(var,"clients")) return dumpConsumerList(); 
    if (strstr(var,"contracts")) return dumpContractList();
    if (strstr(var,"contract"))
    {
      if (index > 0) i = index;
      else if (s != NULL) i = atoi(s);
      else i = 0;
      dumpContractStats(i);
      return 0;
    }
    if (strstr(var,"modules")) return dumpEqmList();
    if (strstr(var,"properties")) return dumpEqmProperties(arg);
    if (strstr(var,"devices")) return dumpEqmDevices(arg);
  }
  if (nConnectionTableEntries)
  {
    if (strstr(var,"connections"))
    {
      dumpConnectionList();      /* non-clients must cover this routine */
      ttyoutput("\n");
      return 0;
    }
    if (strstr(var,"connection"))
    {
      if (index > 0) i = index;
      else if (s != NULL) i = atoi(s);
      else i = 0;
      dumpConnectionStats(i);      
      return 0;
    }
  }
  if (strstr(var,"BurstLimit"))
  {
    sprintf(strbuf,"Burst Limit set at %d",gBurstLimit);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
    sprintf(strbuf,"Cycle Delay set at %d",gCycleDelay);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"stats")) return dumpServerStats();
  sprintf(strbuf,"Variable unavailable: %s",var); ttyoutput(strbuf);
  return -1;
}
int setVal(char *var,char *val)
{
  USRCMD *uc = NULL;
  char *c,*cp=NULL,sstr[64],vstr[32],strbuf[CMDSIZE];
  int i,len,ival[4];
  int v;

  if (var == NULL || val == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_WRITE)
    {
      if (uc->iparam != NULL) 
      {
        *uc->iparam = atoi(val);
        sprintf(strbuf,"%s set to %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        *uc->fparam = (float)strtod(val,NULL);
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = (int)strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
          if (i<4) ival[i] = atoi(val);
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;
      }
    }
  }
  if (strstr(var,"BurstLimit"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetBurstLimit(v);
     sprintf(strbuf,"Burst Limit set to %d",gBurstLimit); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetCycleDelay(v);
     sprintf(strbuf,"Cycle Delay set to %d",gCycleDelay); ttyoutput(strbuf);
     return 0;
  }
# ifdef DOS
  if (strstr(var,"watchdog"))
  {
     if ((v=atoi(val)) == 0) disableWatchDog();
     else enableWatchDog();
     return 0;
  }
# endif
  if (strstr(var,"debug"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     tineDebug = v;
     sprintf(strbuf,"Debug level %d",tineDebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"logdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     if (v > 0)
     {
       if ((dbgfp=fopen("debug.log","w")) == NULL)
         sprintf(strbuf,"Couldn't open debug log file");
       else sprintf(strbuf,"debug logging ON");
       ttyoutput(strbuf);
     }
     else
     {
       if (dbgfp) fclose(dbgfp);
       dbgfp = NULL;
       sprintf(strbuf,"debug logging OFF");
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"almdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     ALMdebug = v;
     sprintf(strbuf,"Alarm Debug level %d",ALMdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     CommandLevel = v;
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"filter"))
  {
     memset(tagNameFilter,0,64);
     if ((c=strchr(val,'\n')) != NULL) *c = 0;
     if (strlen(val) > 0) strncpy(tagNameFilter,val,63);
     ttyoutput("debug name filter %s entered",tagNameFilter);
     return 0;
  }
  sprintf(strbuf,"Variable %s unavailable",var); ttyoutput(strbuf);
  return -1;
}
int dbglog(char *text,...)
{
  static int inside = FALSE;
# ifndef NO_SERVICE_PIPE
# if defined(WIN32)
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  int rc,len=0;
# elif defined(UNIX)
  int i;
# endif
# endif
  int cc=0;
  char str[256];
  va_list args;

  if (inside) return 0;
  inside = TRUE;  /* TODO: use a real mutex */
# if defined(UNIX)
  va_start(args,text);
  vsnprintf(str,256,text,args); str[256] = 0;
# elif defined(WIN32)
  va_start(args,text);
  _vsnprintf(str,256,text,args); str[256] = 0;
# else
  if (strlen(text) > 128) return string_too_long;
  va_start(args,text);
  vsprintf(str,text,args);
# endif
  if (str[strlen(str)-1] == '\n') strcat(str,">");
  else sprintf(&str[strlen(str)]," @%s\n>",getDataTimeString((double)time(NULL),FALSE));
  va_end(args);
  if (strlen(tagNameFilter) && !strstr(str,tagNameFilter)) ccerr(0);
  printf(str);
  if (dbgfp != NULL) fprintf(dbgfp,str);
# ifndef NO_SERVICE_PIPE
# if defined(WIN32)
  if (IPCSocket != 0) 
  {
    rc = WriteFile(pipeInst->hPipeInst,str,(UINT32)strlen(str),&len,NULL);
    if (rc == 0) 
    {
      rc = GetLastError();
      if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
    }
  }
# elif defined(UNIX)
  for (i=0; i<nipcclients; i++) write(IPCfds[i],str,strlen(str));
# endif
# endif
err:
  inside = FALSE;
  return cc;
}

TINE_EXPORT int ttyoutput(char *text,...)
{
  int cc=0;
  char str[512];
  va_list args;
# if defined(UNIX)
  va_start(args,text);
  vsnprintf(str,256,text,args); str[256] = 0;
  va_end(args);
# elif defined(WIN32)
  va_start(args,text);
  _vsnprintf(str,256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) < 128)
  {
    va_start(args,text);
    vsprintf(str,text,args);
    va_end(args);
  }
  else
  {
    strncpy(str,text,256); str[256] = 0;
  }
# endif
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  strcat(str,">");

  printf(str);
  ipcprintf(str); 
  return cc;
}

#endif /* NO_DOXYGEN */
TINE_EXPORT int RegisterUserCommand(
    char *cmd,
    int (*fcn)(int,int,int,int),
    int *iparam,
    float *fparam,
    int access)
{
  USRCMD *uc;

  if ((uc=(USRCMD *)SystemCalloc(1,sizeof(USRCMD))) == NULL) return out_of_local_memory;
  if (access == 0) access = CA_READ;
  strncpy(uc->cmd,cmd,32);
  uc->fcn = fcn;
  uc->iparam = iparam;
  uc->fparam = fparam;
  uc->access = access;
  uc->next = usrcmd;
  usrcmd = uc;
  numUsrCmd++;
  return 0;
}
TINE_EXPORT int RegisterUserFunction(
    char *name,
    int (*fcn)(int argc,char *argv[]))
{
  USRFCN *uf;

  if ((uf=(USRFCN *)SystemCalloc(1,sizeof(USRFCN))) == NULL) return out_of_local_memory;
  strncpy(uf->name,name,32);
  uf->fcn = fcn;
  uf->next = usrfcn;
  usrfcn = uf;
  numUsrFcn++;
  return 0;
}
